# -*- coding: utf-8 -*-
"""
Created on Fri Aug  2 13:55:33 2019

@author: psakicki
"""

import copy
import datetime as dt
#### Import the logger
import logging
import os
########## BEGIN IMPORT ##########
#### External modules
from collections import Counter

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
from matplotlib.widgets import MultiCursor

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import reffram
from geodezyx import stats
from geodezyx import time_series
from geodezyx import utils

log = logging.getLogger('geodezyx')
log.setLevel(logging.INFO)

##########  END IMPORT  ##########

class Point():
      
    def __init__(self,A=0.,B=0.,C=0.,T=0.,initype='XYZ',
                 sA=0.,sB=0.,sC=0.,name='noname',anex=None):
        """
        Initialize a Point Object by importing the coordinate component

        Parameters
        ----------
        A : float
            X, F (latitude), E (depends on initype).
        B : float
            Y, L (longitude), N (depends on initype).
        C : float
            Z, H (hight), U (depends on initype).
        T : float, optional
            Epoch of the measure. The default is 0..
        initype : str, optional
            The inital coordinates type. The default is 'XYZ'.
            The others are 'FLH', 'ENU' and 'NED'
        sA : TYPE, optional
            sigma of A component. The default is 0.
        sB : TYPE, optional
            sigma of B component. The default is 0.
        sC : TYPE, optional
            sigma of C component. The default is 0.
        name : str, optional
            Flexible name for the Point identification. The default is 'noname'.
        anex : dict, optional
            Additional data. The default is None. See Note
        
        Note
        ----
        
        A dictionary called anex is also initialized to allow a
        versatile storage of a variety of data
        
        Exemple of dictionary keys 
        RMS: average RMS (for gipsy)
        sdAB , sdBC , sdAC : the variances between A,B,C (for rtklib)
        sdXY , sdXZ , sdYZ : the variances between XYZ (pbo.pos)
        Vx , Vy , Vz , sVx , sVy , sVz : velocity of the point (EPOS coordinates)
        """
        
        self.Tset(T)
        self.ENUset()
        self.name = name
        if not anex:
            self.anex = dict()
        else:
            self.anex = anex

        # le dico "anex" permet de stocker de manière polyvalente des données diverses
        # On trouvera (LISTE SE DEVANT ETRE LA PLUS EXHAUSTIVE POSSIBLE )
        #
        # RMS : moyenne RMS (pour les gipsy bosser)
        # sdAB , sdBC , sdAC : les variances entre A,B,C (pour les rtklib)
        # sdXY , sdXZ , sdYZ : les variances entre XYZ (pbo.pos)
        # Vx , Vy , Vz , sVx , sVy , sVz : velocity of the point (EPOS coordinates)

        if initype == 'XYZ':
            self.XYZset(A,B,C,sA,sB,sC)
        elif initype == 'FLH': # On travaille en degres decimaux
            self.FLHset(A,B,C,sA,sB,sC)

        elif initype == 'ENU':
            self.ENUset(A,B,C,sA,sB,sC)

        elif initype == 'NED':
            self.NEDset(A,B,C,sA,sB,sC)
            
        elif initype == 'UTM':
            self.UTMset(A,B,C,sA,sB,sC)
        else:
            log.error("wrong initype")

    def __call__(self):

        if self.initype == 'XYZ':
            return self.X,self.Y,self.Z,self.Tdt,self.T
        elif self.initype == 'FLH':
            return self.F,self.L,self.H,self.Tdt,self.T
        elif self.initype == 'ENU':
            return self.E,self.N,self.U,self.Tdt,self.T
        elif self.initype == 'NED':
            return self.N,self.E,self.D,self.Tdt,self.T
        elif self.initype == 'UTM':
            return self.Eutm,self.Nutm,self.Uutm,self.Tdt,self.T
        else:
            log.error("wrong initype")

    def __repr__(self):
        if (not hasattr(self,'X')):
            return "{},{},{},{},{}".format(self.E,self.N,self.U,self.Tdt,self.T)
        else:
            return "{},{},{},{},{}".format(self.X,self.Y,self.Z,self.Tdt,self.T)

    def XYZset(self,X=0,Y=0,Z=0,sX=0,sY=0,sZ=0):
        self.X = X
        self.Y = Y
        self.Z = Z
        self.sX = sX
        self.sY = sY
        self.sZ = sZ

        self.initype = 'XYZ'
        self.F,self.L,self.H = conv.xyz2geo(self.X, self.Y, self.Z)

    def FLHset(self,F=0,L=0,H=0,sF=0,sL=0,sH=0):
        self.F = F
        self.L = L
        self.H = H
        self.sF = sF
        self.sL = sL
        self.sH = sH

        self.initype = 'FLH'
        self.X,self.Y,self.Z = conv.geo2xyz(self.F, self.L, self.H)
        self.sX,self.sY,self.sZ = conv.sigma_geo2xyz(F, L, H, sF, sL, sH)
        

    def ENUset(self,E=np.nan,N=np.nan,U=np.nan,
               sE=np.nan,sN=np.nan,sU=np.nan):
        self.E = E
        self.N = N
        self.U = U
        self.sE = sE
        self.sN = sN
        self.sU = sU

        self.initype = 'ENU'

    def NEDset(self,N=np.nan,E=np.nan,D=np.nan,
               sN=np.nan,sE=np.nan,sD=np.nan):
        self.N = N
        self.E = E
        self.D = D
        self.sN = sN
        self.sE = sE
        self.sD = sD

        self.initype = 'NED'


    def UTMset(self,Eutm=np.nan,Nutm=np.nan,Uutm=np.nan,
               sEutm=np.nan,sNutm=np.nan,sUutm=np.nan):
        self.Eutm = Eutm
        self.Nutm = Nutm
        self.Uutm = Uutm
        self.sEutm = sEutm
        self.sNutm = sNutm
        self.sUutm = sUutm

        self.initype = 'UTM'
        

    def add_offset(self,dA,dB,dC):
        log.warning("add_offset as method are hazardous ...")
        temp = time_series.add_offset_point(self,dA,dB,dC)
        self.__dict__ = temp.__dict__

    def Tset(self,T=0):
        if T == 0:
            T = dt.datetime.now()

        if type(T) == dt.datetime:
            self.Tdt = T
            self.T = conv.dt2posix(T)

        else:
            self.T = float(T)
            self.Tdt = conv.posix2dt(float(T))

    def ENUcalc_pt(self,refENU ):

        self.refENU = refENU

        dX =  self.X - refENU.X
        dY =  self.Y - refENU.Y
        dZ =  self.Z - refENU.Z

        Etmp,Ntmp,Utmp = conv.xyz2enu(self.X, self.Y, self.Z,
                                      refENU.X, refENU.Y, refENU.Z)
        
        self.E,self.N,self.U = Etmp[0],Ntmp[0],Utmp[0]

        if self.initype == 'FLH' and hasattr(self,'s_f'):
            if not np.isnan(self.sF):
                self.sE,self.sN,self.sU = conv.sigma_geo2enu(self.F, self.L, self.H,
                                                             self.sF, self.sL, self.sH)
        elif self.initype == 'XYZ' and hasattr(self,'sX'):
            if np.isnan(self.sX):
                
                return
                
            if 'sdXY' in self.anex.keys():
                sXY = self.anex['sdXY']
                sXZ = self.anex['sdXZ']
                sYZ = self.anex['sdYZ']
            else:
                sXY, sXZ, sYZ = 0,0,0
                
            self.sE,self.sN,self.sU = conv.sigma_xyz2enu(self.X, self.Y, self.Z,
                                                         self.sX, self.sY, self.sZ,
                                                         s_xy=sXY, s_yz=sYZ, s_xz=sXZ)


    def UTMcalc_pt(self,ellips="wgs84"):
        self.Eutm , self.Nutm , _ = conv.utm_geo2xy(self.F,self.L)
        self.Uutm = self.H
        

    def keysanex(self):
        return list(self.anex.keys())


    def helmert_trans(self,params='itrf2008_2_etrf2000',invert=False):
        Xb = reffram.helmert_trans(np.array([self.X,self.Y,self.Z]),params,invert)
        self.XYZset(*Xb)
        return None

    def velocity_trans(self, vx, vy, vz, epoc_init = 'auto' , epoc_end = 'auto'):
        """
        auto == epoc of the measures
        """

        if epoc_init == 'auto' and epoc_end == 'auto':
            log.error('epoc_init == auto and epoc_end == auto')
            return None

        tdt = conv.posix2dt(self.T)
        yeardec = conv.dt2year_decimal(tdt)

        if epoc_init  == 'auto':
            epoc_init = yeardec

        if epoc_end  == "auto":
            epoc_end = yeardec

        Xb = reffram.itrf_speed_calc(self.X,self.Y,self.Z, epoc_init ,
                                  vx, vy, vz, epoc_end)
        self.XYZset(*Xb)
        return None

class TimeSeriePoint:
    
    def __init__(self,stat='STAT'):
        """
        Initialize a TimeSeriePoint object

        Parameters
        ----------
        stat : str, optional
            station 4-char. code. The default is 'STAT'.

        """

        self.pts = []
        self.i_nomi = 0

        self.meta_set(stat=stat)

        self.boolENU = False
        self.boolUTM = False

        self.bool_interp_uptodate = False
        self.bool_discont = False
        self.bool_discont_manu = False

        self.anex = dict()

    def __repr__(self):
        """
        Representation of a TimeSeriePoint object
        """
        if self.pts == []:
            raise Exception('ERR: TimeSeriePoint is empty ...')

        start = self.startdate()
        end = self.enddate()
        nbday = int((end - start).days + 1.)

        ratio = self.nbpts * 100. / nbday

        return '{} {} {} {} {} {} {} {:5.2f}{}'.format(self.stat,self.nbpts,'points',
                start,end,nbday,"nb days", ratio,"%")

    def __getitem__(self,i):
        return self.pts[i]

    @property
    def nbpts(self):
        """
        Method to have the length of the TimeSerie

        Returns
        -------
        int
            Length of the TimeSerie.

        """
        return len(self.pts)


    def meta_set(self,path='',stat='STAT',name=''):
        """
        Set meta data about the TimeSerie

        Parameters
        ----------
        path : str, optional
            file path. The default is ''.
        stat : str, optional
            station 4-char. code. The default is 'STAT'.
        name : str, optional
            free name of for the TS, 
            like the experience, the periode , the software ...
            The default is ''.

        Returns
        -------
        None.
        """

        self.path = path
        self.stat = stat

        bn = os.path.basename(path)
        dn = os.path.dirname(path)

        if name == '' :
            if bn == 'tdp_final':
                self.name = os.path.basename(dn)
            else:
                self.name = bn
        else:
            self.name = name
        self.interval_nominal()

    def del_data(self):
        """
        Method to purge the data in the TimeSeriePoint

        Returns
        -------
        None.

        """
        self.pts = []

        self.bool_interp_uptodate = False
        self.bool_discont = False


    def readfile(self,filein):
        """
        Method to read the data form a file
        Should be used with care

        Parameters
        ----------
        filein : str
            path of the file.

        Returns
        -------
        None.

        """
        self.__dict__ = files_rw.read_all_points(filein).__dict__

        self.interp_set()

    def add_point(self, point_inp):
        """
        Method to add a Point in the TimeSerie Object

        Parameters
        ----------
        point_inp : Point Object

        Returns
        -------
        None.

        """
        self.pts.append(point_inp)
        # this line is discontiued, because now nbpts is a property
        #self.nbpts = len(self.pts)

        self.bool_interp_uptodate = False

    def aleapt(self):
        """
        Method to get a random Point in the TimeSeries

        Returns
        -------
        Point Object

        """
        ipt = np.random.randint(self.nbpts)
        log.info("random point selected #%s",str(ipt))

        return self.pts[ipt]

    def startdate(self):
        """
        Method to get the first epoch of the data in the TimeSerie

        Returns
        -------
        DateTime

        """
        self.sort()
        return conv.posix2dt(self.pts[0].T)

    def enddate(self):
        """
        Method to get the last epoch of the data in the TimeSerie

        Returns
        -------
        DateTime

        """
        self.sort()
        return conv.posix2dt(self.pts[-1].T)
    
    
    def len_period(self,output_seconds=False):
        """
        Method to get the period length

        Returns
        -------
        timedelta or 
        """    
        delta = self.enddate() - self.startdate()
        
        if output_seconds:
            return delta.days*86400 + delta.seconds
        else:
            return delta

    def interval_nominal(self):
        """
        Method to get the nominal internal between two epochs.

        Returns
        -------
        float
            interval nominal.

        """

        if len(self.pts) < 2:
            self.i_nomi = 0
        else:
            Ttemp = np.sort([pt.T for pt in self.pts])
            self.i_nomi = np.round(np.min(np.diff(Ttemp)),1)

        return self.i_nomi


    def from_list(self,T,A,B,C,coortype='XYZ',sA=[],sB=[],sC=[]):
        """
        Method to load data from lists to the TimeSerie

        Parameters
        ----------
        T : float
            Time.
        A : list of float
            X, F (latitude), E..
        B : list of float
            Y, L (longitude), N.
        C : list of float
            Z, H (hight), U.
        coortype : str, optional
            The coordinates type. The default is 'XYZ'.
        sA : list of float, optional
            sigma of A component. The default is [].
        sB : list of float, optional
            sigma of B component. The default is [].
        sC : list of float, optional
            sigma of C component. The default is [].

        Returns
        -------
        None.

        """

        if not sA:
            sA = np.zeros(len(A))
        if not sB:
            sB = np.zeros(len(B))
        if not sC:
            sC = np.zeros(len(C))

        for t,a,b,c,sa,sb,sc in zip(T,A,B,C,sA,sB,sC):
            point = Point(a,b,c,t,coortype,sa,sb,sc)

            self.add_point(point)

        if coortype == "ENU":
            self.boolENU = True

        if coortype == "UTM":
            self.boolUTM = True

        self.sort()

        return None


    def to_list(self,coortype='XYZ',specific_output=None,
                time_as_datetime=False):
        """
        Export the TimeSerie Object as Lists (Numpy Arrays)

        Parameters
        ----------
        coortype : str, optional
        The coordinates type exported to the list.
        The default is 'XYZ'.
        specific_output : int, optional
            ask for a specific list, ranges between 0 and 6.
            The default is None.
        time_as_datetime : bool, optional
            if True the Time list is exported in datetime
            if False the Time list is exported in Posix time

        Returns
        -------
        A,B,C,T,sA,sB,sC : lists
            A = X, F (latitude), E.
            B = Y, L (longitude), N.
            C = Z, H (hight), U.
            T = Time
            sA = sigma of A component
            sB = sigma of B component
            sC = sigma of C component
        """


        if coortype == 'XYZ':
            A,B,C = 'X','Y','Z'
            sA,sB,sC = 'sX','sY','sZ'

        elif coortype == 'FLH':
            A,B,C = 'F','L','H'
            sA,sB,sC = 'sF','sL','sH'

        elif coortype == 'ENU':
            if not self.boolENU:
                log.warning("no ENU coord. for " + self.name)
                return None

            A,B,C = 'E','N','U'
            sA,sB,sC = 'sE','sN','sU'

        elif coortype == 'UTM':
            if not self.boolUTM:
                log.warning("no UTM coord. for " + self.name)
                return None

            A,B,C = 'Eutm','Nutm','Uutm'
            sA,sB,sC = 'sEutm','sNutm','sUutm'

        else:
            log.error("coortype does not exist")

        if self.nbpts == 0:
            log.error(self.name + " the timeserie is empty")

        A = np.asarray([ getattr(pt,A) for pt in self.pts ])
        B = np.asarray([ getattr(pt,B) for pt in self.pts ])
        C = np.asarray([ getattr(pt,C) for pt in self.pts ])
        T = np.asarray([ pt.T for pt in self.pts ])

        if hasattr(self.pts[0],sA):
            sA = np.asarray([ getattr(pt,sA) for pt in self.pts ])
            sB = np.asarray([ getattr(pt,sB) for pt in self.pts ])
            sC = np.asarray([ getattr(pt,sC) for pt in self.pts ])
        else:
            sA = np.asarray([ np.nan ] * len(self.pts))
            sB = np.asarray([ np.nan ] * len(self.pts))
            sC = np.asarray([ np.nan ] * len(self.pts))
        # il faut squeezer les vecteurs parce que des fois on se retrouve
        # avec des matrices
        
        if time_as_datetime:
            Tout = conv.posix2dt(T)
        else:
            Tout = T
        
        sq = np.squeeze
        outtup = (sq(A) , sq(B) , sq(C) , sq(Tout) , sq(sA) , sq(sB) , sq(sC))
        if specific_output == None:
            return outtup
        elif type(specific_output) is int:
            return outtup[specific_output]
        else:
            log.info("this mode must be implemented ;)")
            log.info("use an int as index instead")
            return outtup
        
    def to_dataframe(self,coortype ='XYZ'):
        """
        Export the TimeSerie Object as DataFrame
        
        Parameters
        ----------
        coortype : str or iterable of str.
            The coordinates type exported to the DataFrame.
            'XYZ', 'FLH', 'ENU', 'NED'
            can be also an iterable like ('XYZ','FLH')
            The default is 'XYZ'.

        Returns
        -------
        DF : DataFrame
            output DataFrame.
        """
        
        if not utils.is_iterable(coortype):
            coortype = (coortype,)
        
        ColStk = tuple()
        ColNameStk = []
        
        for icoty , coty in enumerate(coortype):
            A,B,C,T,sA,sB,sC = self.to_list(coty)
            
            if coty == "UTM":
                cotycolnam = ["Eutm","Nutm","Uutm"]
            else:
                cotycolnam = coty
                
            if icoty == 0:
                Tdt = conv.posix2dt(T)
                ColStk = ColStk + (Tdt,T,A,B,C,sA,sB,sC)  
                ColNameStk = ["Tdt","T"] + [e for e in cotycolnam] + ["s" + e for e in cotycolnam]
            else:
                ColStk = ColStk + (A,B,C,sA,sB,sC)
                ColNameStk = [e for e in coty] + ["s" + e for e in coty]
                
        BIG = np.column_stack(ColStk)
        DF = pd.DataFrame(BIG)
        DF.columns = ColNameStk
        DF = DF.infer_objects()
        
        return DF

    def sort(self):
        """
        Internal method to sort the point in the TimeSerie Object

        Returns
        -------
        None.

        """
        self.pts.sort(key=lambda x: x.T)

    def plot(self,coortype='ENU',
             diapt=2,
             alpha=0.8,
             fig=1,
             errbar=True,
             new_style=True,
             symbol = '.',
             errbar_width=1,
             ylim=None):
        """
        Plot data in a TimeSerie Object

        Parameters
        ----------
        coortype : str, optional
            The coordinates type. The default is 'ENU'.
        diapt : float, optional
            Point diamaeter. The default is 2.
        alpha : float, optional
            Alpha (transparency) of points. The default is 0.8.
        fig : int or Figure object, optional
            Figure ID where the data will be plotted
            can accept a int (id of a Figure)
            OR the figure Object itself.
            The default is 1.
        errbar : bool, optional
            Plot the error bars. The default is True.
        new_style : bool, optional
            Plot in a new style.
            The old style is only kept for legacy
            The default is True.
        symbol : str, optional
            symbol. The default is '.'.
        errbar_width : TYPE, optional
            coefficient for the error bar size. The default is 1.

        Returns
        -------
        The matplotlib Figure object.

        """
        
        log.setLevel(logging.INFO)

        if new_style:
            styleint = 310
        else:
            styleint = 220

        try:
            A, B, C , T , sA, sB, sC = self.to_list(coortype=coortype)
        except TypeError as tyer:
            log.error("unable to get coordinates")
            log.info("TRICK : check if the given coortype is in the timeserie")
            raise tyer

        if coortype == 'ENU':
            Atitle = 'East'
            Btitle = 'North'
            Ctitle = 'Up'
            ABtitle = 'East North'
            yylabel = 'displacement (m)'

        elif coortype == 'XYZ':
            Atitle = 'X'
            Btitle = 'Y'
            Ctitle = 'Z'
            yylabel = 'displacement (m)'
            ABtitle = 'X Y (sans signification)'

        elif coortype == 'FLH':
            Atitle = 'Phi'
            Btitle = 'Lambda'
            Ctitle = 'Haut'
            yylabel = 'displacement (m)'
            ABtitle = 'Phi Lambda (sans signification)'


        elif coortype == 'UTM':
            Atitle = 'East (UTM)'
            Btitle = 'North (UTM)'
            Ctitle = 'Up'
            yylabel = 'displacement (m)'
            ABtitle = 'East North (UTM)'

        else:
            Atitle = 'A'
            Btitle = 'B'
            Ctitle = 'C'
            yylabel = 'displacement (??)'
            ABtitle = 'A & B'

        log.info("plot : %s, pts : %s", self.nbpts, self.stat)

        namest=0
        namend=10

        Tdt = conv.posix2dt(T)

        if type(fig) is int:
            figobj = plt.figure(fig)
        elif type(fig) is plt.Figure:
            figobj = fig

        figobj.suptitle(self.stat)


        if self.name:
            name4plot = self.name[namest:namend]
        else:
            name4plot = self.stat

        plt.subplot(styleint+1)
        
        if errbar:
            plt.errorbar(Tdt,A,sA, fmt=symbol,label=name4plot,
                         markersize=diapt, alpha=alpha,ecolor='xkcd:light grey',
                         elinewidth=errbar_width)
        else:
            plt.plot(Tdt,A,symbol,label=name4plot,
                     markersize=diapt, alpha=alpha)
        try:
            plt.legend()
        except:
            pass
        #plt.xlabel('Date')
        plt.ylabel(yylabel)
        plt.title(Btitle)
        plt.title(Atitle)
                
#        ax = plt.gca()

#        if coortype == 'ENU':
##            refstr = 'ref XYZ = ' + utils.join_improved(',',self.refENU.X ,
##                                                          self.refENU.Y ,
##                                                          self.refENU.Z)
        plt.subplot(styleint+2)
        if errbar:
            plt.errorbar(Tdt,B,sB, fmt=symbol,label=name4plot,
                         markersize=diapt, alpha=alpha,ecolor='xkcd:light grey',
                         elinewidth=errbar_width)
        else:
            plt.plot(Tdt,B,symbol,label=name4plot,
                     markersize=diapt, alpha=alpha)
        try:
            plt.legend()
        except:
            pass
        #plt.xlabel('Date')
        plt.ylabel(yylabel)
        plt.title(Btitle)

        plt.subplot(styleint+3)
        if errbar:
            plt.errorbar(Tdt,C,sC, fmt=symbol,label=name4plot,
                         markersize=diapt, alpha=alpha,ecolor='xkcd:light grey',
                         elinewidth=errbar_width)
        else:
            plt.plot(Tdt,C,symbol,label=name4plot,
                     markersize=diapt, alpha=alpha)
        try:
            plt.legend()
        except:
            pass
        
        if ylim:
            [a.set_ylim(ylim) for a in figobj.axes]
        
        plt.xlabel('Date')
        plt.ylabel(yylabel)
        plt.title(Ctitle)
        figobj.autofmt_xdate()
        figobj.set_size_inches(8.27,11.69)
        figobj.tight_layout()
        plt.subplots_adjust(top=0.93)

        if not new_style:
            plt.subplot(styleint+4)
            plt.axis('equal')
            try:
                plt.legend()
            except:
                pass
            plt.plot(A,B,'.',label=name4plot,
                     markersize=diapt, alpha=alpha)
            plt.xlabel(Atitle + ' ' + yylabel)
            plt.ylabel(Btitle + ' ' + yylabel)
            plt.title(ABtitle)

        return figobj


    def plot_discont(self,fig=1):
        """
        Plot discontinuties of a TimeSerie Object contained in discont list

        Parameters
        ----------
        fig : int or Figure object, optional
            Figure ID where the data will be plotted
            can accept a int (id of a Figure)
            OR the figure Object itself.
            The default is 1.

        Returns
        -------
        None.

        """

        if not self.bool_discont:
            log.warning("no discontinuities in TimeSerie")
            return None

        if type(fig) is int:
            figobj = plt.figure(fig)
        elif type(fig) is plt.Figure:
            figobj = fig

        for ax in figobj.axes:
            stats.plot_vertical_bar_ax(self.discont,ax,"r")

        if self.bool_discont_manu:
            for ax in figobj.axes:
                stats.plot_vertical_bar_ax(self.discont_manu,ax,"g")

#        figobj.axes[1]
#        stats.plot_vertical_bar(self.discont)
#
#        figobj.axes[2]
#        stats.plot_vertical_bar(self.discont)


    def discont_manu_click(self,fig=1):
        """
        manual discontinuities are both recorded in the "main" discont list
        and in a new discont_manu list,
        thus the manual discontinuites can be identified

        IMPORTANT : cursor objects (multi , cid)
                    must be stored as global variables like this :
                    multi , cid = tsout.discont_manu_click()


        NOTE : This method was created before point_n_click_plot():
            this other one is more complete
            both has to be merged ASAP !!!!!

        """

        if type(fig) is int:
            figobj = plt.figure(fig)
        elif type(fig) is plt.Figure:
            figobj = fig

        if not self.bool_discont:
            self.discont = []

        if not self.bool_discont_manu:
            self.discont_manu = []

        log.info("press SPACE to record a manual discontinuity")


        def onclick_discont(event):
            ix, iy = matplotlib.dates.num2date(event.xdata).replace(tzinfo=None), event.ydata

            log.info("discontinuity recorded : %s" , ix)
            for ax in figobj.axes:
                stats.plot_vertical_bar_ax([ix],ax,"g")
#

            self.bool_discont      = True
            self.bool_discont_manu = True

            self.discont.append(ix)
            self.discont = sorted(self.discont)

            self.discont_manu.append(ix)
            self.discont_manu = sorted(self.discont_manu)

            #figobj.show()
            plt.draw()

            return None

        multi = MultiCursor(figobj.canvas, figobj.axes , color='k', lw=1)
        cid   = figobj.canvas.mpl_connect('key_press_event', onclick_discont)

        return multi , cid

    def initype(self):
        L = [ pt.initype for pt in self.pts ]
        return Counter(L).most_common(1)[0][0]

    def ENUcalc(self, refENU ):
        """
        Method to determine the ENU components based on a reference point

        Parameters
        ----------
        refENU : Point Object or TimeSeriePoint Object
            Reference point.

        Returns
        -------
        None.

        """
        if refENU.__class__.__name__ == 'Point':
            self.refENU = refENU
            [ pt.ENUcalc_pt(refENU) for pt in self.pts ]
            self.boolENU = True
            self.bool_interp_uptodate = False
            self.interp_set()
        elif refENU.__class__.__name__ == 'TimeSeriePoint':
            self.refENU = refENU.mean_posi()
            refENU.interp_set()
            [ pt.ENUcalc_pt(Point(A=refENU.XfT(pt.T),
                                  B=refENU.YfT(pt.T),
                                  C=refENU.ZfT(pt.T),
                                  initype='XYZ',
                                  T=pt.T)) for pt in self.pts ]

            self.boolENU = True
            self.bool_interp_uptodate = False
            self.interp_set()

    def ENUcalc_from_mean_posi(self,mean_type="median"):
        """
        Method to determine the ENU components based directly 
        on the mean/median position

        Returns
        -------
        None.

        """
        self.ENUcalc(self.mean_posi(mean_type="median"))
        return None
    
    
    def ENUcalc_from_first_posi(self):
        """
        Method to determine the ENU components based directly
        on the mean/median position

        Returns
        -------
        None.

        """
        self.ENUcalc(self.pts[0])
        return None

    def from_uniq_point(self,Point,startdate,enddate,pas=1):
        self.del_data()

        if type(startdate) == dt.datetime:
            startdate = conv.dt2posix(startdate)
        if type(enddate) == dt.datetime:
            enddate = conv.dt2posix(enddate)

        N = int(np.round( enddate - startdate / pas ))
        #datelist = np.arange(startdate,enddate,pas)

        Point.Tset(startdate)

        for i in range(N):
            Point.Tset(Point.T + pas)
            self.add_point(copy.copy(Point))

        self.i_nomi = pas
        
        
    def UTMcalc(self):
        """
        Method to determine the UTM E and N projected coordinates

        Parameters
        ----------

        Returns
        -------
        None.

        """
        self.boolUTM = True
        [ pt.UTMcalc_pt() for pt in self.pts ]
        

    def time_win(self,windows,mode='keep'):
        '''IL EST TRES DANGEREUX DE L'APPLIQUER UN FENETRAGE A SOI MEME'''
        self.__dict__ = time_series.time_win(self,windows,mode).__dict__


    def interp_set(self,interptype = 'slinear'):
        """
        Method to set the coordinate interpolators

        Parameters
        ----------
        interptype : TYPE, optional
            Interpolation type. The default is 'slinear'.

        Returns
        -------
        None.

        """
        if (not hasattr(self.pts[0],'E')) or np.isnan(self.pts[0].E) == True:
            log.warning("no ENU for " + self.name)
        else:
            E,N,U,T,_,_,_ = self.to_list('ENU')
            self.EfT = scipy.interpolate.interp1d(T,E,bounds_error=False,kind=interptype)
            self.NfT = scipy.interpolate.interp1d(T,N,bounds_error=False,kind=interptype)
            self.UfT = scipy.interpolate.interp1d(T,U,bounds_error=False,kind=interptype)

        if (not hasattr(self.pts[0],'X')) or np.isnan(self.pts[0].X) == True:
            log.warning("no XYZ for " + self.name)
        else:

            X,Y,Z,T,_,_,_ = self.to_list('XYZ')

            self.XfT = scipy.interpolate.interp1d(T,X,bounds_error=False,kind=interptype)
            self.YfT = scipy.interpolate.interp1d(T,Y,bounds_error=False,kind=interptype)
            self.ZfT = scipy.interpolate.interp1d(T,Z,bounds_error=False,kind=interptype)


        if (not hasattr(self.pts[0],'F')) or np.isnan(self.pts[0].L) == True:
            log.warning("no FLH for " + self.name)
        else:
            F,L,H,T,_,_,_ = self.to_list('FLH')

            self.FfT = scipy.interpolate.interp1d(T,F,bounds_error=False,kind=interptype)
            self.LfT = scipy.interpolate.interp1d(T,L,bounds_error=False,kind=interptype)
            self.HfT = scipy.interpolate.interp1d(T,H,bounds_error=False,kind=interptype)

        if (not hasattr(self.pts[0],'Eutm')) or np.isnan(self.pts[0].Eutm) == True:
            #log.warning("no UTM for " + self.name)
            pass
        else:
            Eutm,Nutm,Uutm,T,_,_,_ = self.to_list('UTM')

            self.EutmfT = scipy.interpolate.interp1d(T,Eutm,bounds_error=False,kind=interptype)
            self.NutmfT = scipy.interpolate.interp1d(T,Nutm,bounds_error=False,kind=interptype)
            self.UutmfT = scipy.interpolate.interp1d(T,Uutm,bounds_error=False,kind=interptype)

        self.bool_interp_uptodate = True

    def interp_get(self,T,coortype='ENU'):
        """
        Method to get the coordinate interpolators

        Parameters
        ----------
        T : float or list of float
            Time (IN POSIX Time) where the interpolation is wished.
        coortype : str, optional
            The coordinates type. The default is 'ENU'.

        Returns
        -------
        tsout : 
            DESCRIPTION.

        """

        if self.bool_interp_uptodate == False:
            log.warning("interp obsolete, recalcul auto")
            self.interp_set()

        tsout = copy.copy(self)
        tsout.del_data()

        if not utils.is_iterable(T):
            T = np.array([T])

        if coortype == 'ENU':
            A = self.EfT(T)
            B = self.NfT(T)
            C = self.UfT(T)

        if coortype == 'XYZ':
            A = self.XfT(T)
            B = self.YfT(T)
            C = self.ZfT(T)

        if coortype == 'FLH':
            A = self.FfT(T)
            B = self.LfT(T)
            C = self.HfT(T)

        if coortype == 'UTM':
            A = self.EutmfT(T)
            B = self.NutmfT(T)
            C = self.UutmfT(T)

        for i in range(len(T)):
            tsout.add_point(Point(A[i],B[i],C[i],T=T[i],initype=coortype))

        return tsout

    def set_discont(self,indiscont):
        """
        Method to set the discontinuties list

        Parameters
        ----------
        indiscont : list of time
            Discontinuities in the TimeSerie.

        Returns
        -------
        None.

        """
        self.discont = indiscont
        self.bool_discont = True

    def mean_posi(self,coortype='XYZ',outtype='point',mean_type='median'):
        """
        Method to determine the mean position of the TimeSerie

        Parameters
        ----------
        coortype : TYPE, optional
            The coordinates type. The default is 'XYZ'.
        outtype : TYPE, optional
            'point' or 'tuple'. The default is 'point'.
        mean_type : TYPE, optional
            'mean' or 'median'. The default is 'median'.

        Returns
        -------
        Point or coordinates tuple

        """

        #special case where only one point
        if self.nbpts == 1:
            return copy.copy(self)

        A,B,C,T,sA,sB,sC = self.to_list(coortype=coortype)


        if mean_type == 'mean':
            Aout = np.nanmean(A)
            Bout = np.nanmean(B)
            Cout = np.nanmean(C)
        elif mean_type == 'median':
            Aout = np.nanmedian(A)
            Bout = np.nanmedian(B)
            Cout = np.nanmedian(C)

        Tout = (np.max(T) - np.min(T)) / 2

        if outtype == "point":
            out = Point(Aout,Bout,Cout,Tout)
        elif outtype == "tuple":
            out = (Aout,Bout,Cout,Tout)
        else:
            out = (Aout,Bout,Cout,Tout)

        return out

    def add_offset(self,dA,dB,dC):
        """
        NOTE 160415 : add_offset as method are hazardous ...
        use fct add_offset_ts instead
        """
        log.warning("add_offset as method are hazardous ...")
        for pt in self.pts:
            pt.add_offset(dA,dB,dC)

    def decimate(self,dec):
        """
        Method to decimate a TimeSerie

        Parameters
        ----------
        dec : int
            keep 1/dec point in the TimeSerie.

        Returns
        -------
        None.

        """
        self.__dict__ = time_series.decimate_cleaner(self,dec).__dict__
        #decimate_cleaner(self,dec,True)

    def find_point(self,tin,tol=0.001,stop_when_found=True):
        """
        Method to find a specific point according to its timestamp

        Parameters
        ----------
        tin : float or datetime
            timestamp of the researched point.
        tol : float, optional
            tolerence of the research. The default is 0.001.
        stop_when_found : bool, optional
            Stop the research when a point is found. The default is True.

        Returns
        -------
        Point Object
            Point Found.
        int or list of int
            index of the point.

        """

        find = False

        if type(tin) is dt.datetime:
            tin = conv.dt2posix(tin)

        tmin = self.pts[0].T
        tmax  = self.pts[-1].T

        if not ( tmin <= tin <= tmax ):
            log.warning("t !€ [tmin , tmax]")

        if stop_when_found:
            for i,p in enumerate(self.pts):
                if np.abs(p.T - tin) < tol:
                    find = True
                    break
            if find:
                return p,i
            else:
                return Point(),np.nan

        else:
            pts_stk = []
            i_stk = []
            for i,p in enumerate(self.pts):
                if np.abs(p.T - tin) < tol:
                    pts_stk.append(p)
                    i_stk.append(i)
            return pts_stk  , i_stk
        
    def rm_duplicat_pts(self, coortype="XYZ"):
        """
        Remove duplicate points from the time series.

        Parameters
        ----------
        coortype : str, optional
            The coordinate type to consider for duplication check. The default is "XYZ".

        Returns
        -------
        None
        """
        T = self.to_dataframe(coortype)["T"]

        dup_bool = T.duplicated()

        if dup_bool.sum() > 0:
            log.warning("%s duplicated point(s) removed for %s",
                        dup_bool.sum(), self.name)

        self.pts = list(pd.Series(self.pts)[np.logical_not(dup_bool)])



 #  ______                      _                      _        _    _____ _                         
 # |  ____|                    (_)                    | |      | |  / ____| |                        
 # | |__  __  ___ __   ___ _ __ _ _ __ ___   ___ _ __ | |_ __ _| | | |    | | __ _ ___ ___  ___  ___ 
 # |  __| \ \/ / '_ \ / _ \ '__| | '_ ` _ \ / _ \ '_ \| __/ _` | | | |    | |/ _` / __/ __|/ _ \/ __|
 # | |____ >  <| |_) |  __/ |  | | | | | | |  __/ | | | || (_| | | | |____| | (_| \__ \__ \  __/\__ \
 # |______/_/\_\ .__/ \___|_|  |_|_| |_| |_|\___|_| |_|\__\__,_|_|  \_____|_|\__,_|___/___/\___||___/
 #             | |                                                                                   
 #             |_|                                                                                   


    
    
class Attitude:
    def __init__(self,R=0,P=0,Y=0,T=0,sR=0,sP=0,sY=0,devID='NULL',angtype='deg'):

        self.Tset(T)
        self.devID = devID

        if angtype == 'deg':
            self.R = R
            self.P = P
            self.Y = Y

        elif angtype == 'rad' :
            self.R = np.rad2deg(R)
            self.P = np.rad2deg(P)
            self.Y = np.rad2deg(Y)
        else:
            raise Exception("Mauvais angtype")

        self.Qcalc()

    def __call__(self):
        return self.R,self.P,self.Y,self.Tdt

    def __repr__(self):
        return "{},{},{},{}".format(self.R,self.P,self.Y,self.Tdt)


    def Tset(self,T=0):
        if T == 0:
            T = dt.datetime.now()

        if type(T) == dt.datetime:
            self.Tdt = T
            self.T = conv.dt2posix(T)

        else:
            self.T = float(T)
            self.Tdt = conv.posix2dt(float(T))

    def RPYget(self):
        return self.R,self.P,self.Y

    def RPYset(self,R=0,P=0,Y=0,sR=0,sP=0,sY=0):
        self.R  = R
        self.P  = P
        self.Y  = Y
        self.sR  = sR
        self.sP  = sP
        self.sY  = sY

    def Qcalc(self):
        self.Q = conv.quaternion(self.R , self.P , self.Y , 'deg')
        return None


    
class TimeSerieObs(object):

    ''' LES DIFFERENCES AVEC TSPOINT
        * Les objets ne contiennent qu'un type de données sous une seul forme
        (a la difference d'un point qui peut exister sous plusieurs formes)
        * Dans un fichier en input, il peut y avoir plusieurs "devices"
          => les fonctions de lectures produisent donc obligatoirement des listes
          de TS (le cas échéant une liste à 1 élt)
          => la methode readfile() nécessite donc l'indice de la device'''

    def __init__(self,typeobs='NULL',filepath =''):

        self.obs = []
        self.nbobs = 0
        self.i_nomi = 0
        self.typeobs = typeobs
        # ASM Attitude
        self.bool_interp_uptodate = False

        self.meta_set(filepath)

    def meta_set(self,path='',devID='NULL',name=''):
        self.path = path
        self.devID = devID

        bn = os.path.basename(path)
        dn = os.path.dirname(path)

        if name == '' :
            if bn == 'tdp_final':
                self.name = os.path.basename(dn)
            else:
                self.name = bn

        self.interval_nominal()

    def del_data(self):
        self.obs = []
        self.nbobs = 0

    def readfile(self,filein,indtab=0):
        log.info("selection device %s" , indtab)
        temp = files_rw.read_all_obs(filein)[indtab]
        self.__dict__ = temp.__dict__

        self.interp_set()

    def add_obs(self,inObs):
        self.obs.append(inObs)
        self.nbobs = len(self.obs)

        self.bool_interp_uptodate = False

    def aleaobs(self):
        iobs = np.random.randint(0,self.nbobs)
        log.info("observation no " + str(iobs))

        log.info(self.obs[iobs])

        return self.obs[iobs]


    def interval_nominal(self):
        if len(self.obs) < 2:
            self.i_nomi = 0
        else:
            Ttemp = np.sort([o.T for o in self.obs])
            self.i_nomi = np.round(np.min(np.diff(Ttemp)),1)

        return self.i_nomi

    def timewin(self,windows,mode='keep'):
        '''IL EST TRES DANGEREUX DE L'APPLIQUER UN FENETRAGE A SOI MEME'''
        self.__dict__ = time_series.time_win(self,windows,mode).__dict__

    def startdate(self):
        return self.obs[0].T

    def enddate(self):
        return self.obs[-1].T

    def to_list(self):
        if self.typeobs == 'NULL':
            log.error("pas de typeobs defini (NULL)")
            return 0

        if self.typeobs == 'RPY':
            A,B,C = 'R','p','Y'
            sA,sB,sC = 'sR','sP','sY'

        A = np.asarray([ getattr(o,A) for o in self.obs ])
        B = np.asarray([ getattr(o,B) for o in self.obs ])
        C = np.asarray([ getattr(o,C) for o in self.obs ])
        T = np.asarray([ o.T for o in self.obs ])

        if hasattr(self,sA):
            sA = np.asarray([ getattr(o,sA) for o in self.obs ])
            sB = np.asarray([ getattr(o,sB) for o in self.obs ])
            sC = np.asarray([ getattr(o,sC) for o in self.obs ])

        else:
            sA = np.asarray([ np.nan ] * len(self.obs))
            sB = np.asarray([ np.nan ] * len(self.obs))
            sC = np.asarray([ np.nan ] * len(self.obs))

        return A, B, C , T , sA , sB , sC



    def interp_set(self,interptype='slinear'):

        if self.typeobs == 'NULL':
            log.error("no typeobs defined (NULL)")
            return 0

        if self.typeobs == 'RPY':

            R,P,Y,T,_,_,_ = self.to_list()

            self.RfT = scipy.interpolate.interp1d(T,R,bounds_error=False,kind=interptype)
            self.PfT = scipy.interpolate.interp1d(T,P,bounds_error=False,kind=interptype)
            self.YfT = scipy.interpolate.interp1d_ang(T,Y,bounds_error=False,kind=interptype)

        self.bool_interp_uptodate = True

    def interp_get(self,T):

        if self.bool_interp_uptodate == False:
            log.warning("interp obsolete, recalcul auto")
            self.interp_set()

        tsout = copy.copy(self)
        tsout.del_data()

        if not utils.is_iterable(T):
            T = np.array([T])


        if self.typeobs == 'NULL':
            log.error("no typeobs defined (NULL)")
            return 0

        if self.typeobs == 'RPY':
            A = self.RfT(T)
            B = self.PfT(T)
            C = self.YfT(T)

            for i in range(len(T)):
                    tsout.add_obs(Attitude(A[i],B[i],C[i],T=T[i]))

        return tsout

    def plot(self,diapt=10,alpha=0.8,fig=1,new_style=True):

        A, B, C , T , sA, sB, sC = self.to_list()

        log.info("plot : %s, pts : %s", self.nbpts, self.stat)

        if self.typeobs == 'RPY':
            listtitle = ['','Roll','Pitch','Yaw']
        else:
            listtitle = ['','','','']

        namest=0
        namend=10
        Tdt = conv.posix2dt(T)

        if new_style:
            styleint = 410
        else:
            styleint = 220

        plt.figure(fig)
        plt.subplot(styleint + 1)
        plt.axis('equal')
        plt.plot(A,B,'.',label=str(self.devID)[namest:namend], markersize=diapt, alpha=alpha)
        plt.legend()
        plt.title(listtitle[0])

        plt.subplot(styleint + 2)
        plt.plot(Tdt,A,'.',label=str(self.devID)[namest:namend], markersize=diapt, alpha=alpha)
        plt.legend()
        plt.title(listtitle[1])

        plt.subplot(styleint + 3)
        plt.plot(Tdt,B,'.',label=str(self.devID)[namest:namend], markersize=diapt, alpha=alpha)
        plt.legend()
        plt.title(listtitle[2])

        plt.subplot(styleint + 4)
        plt.plot(Tdt,C,'.',label=str(self.devID)[namest:namend], markersize=diapt, alpha=alpha)
        plt.legend()
        plt.title(listtitle[3])


#  _____       _                      _   _             _____  _       _
# |_   _|     | |                    | | (_)           |  __ \| |     | |
#   | |  _ __ | |_ ___ _ __ __ _  ___| |_ ___   _____  | |__) | | ___ | |_
#   | | | '_ \| __/ _ \ '__/ _` |/ __| __| \ \ / / _ \ |  ___/| |/ _ \| __|
#  _| |_| | | | ||  __/ | | (_| | (__| |_| |\ v /  __/ | |    | | (_) | |_
# |_____|_| |_|\__\___|_|  \__,_|\___|\__|_|_\_/ \___| |_|    |_|\___/ \__|
# |  __ \    (_)     | |     ___     / ____| (_)    | |
# | |__) |__  _ _ __ | |_   ( _ )   | |    | |_  ___| | __
# |  ___/ _ \| | '_ \| __|  / _ \/\ | |    | | |/ __| |/ /
# | |  | (_) | | | | | |_  | (_>  < | |____| | | (__|   <
# |_|   \___/|_|_| |_|\__|  \___/\/  \_____|_|_|\___|_|\_\


    """
    USAGE :

        Then :

            PnC = point_n_click_plot()

            multi , cid = PnC(fig=6)

            PnC.selectedX

        i.e. :

            Create an object point_n_click_plot (here it is PnC in the exemple below)

            Call the object like a function with the id of the plot figure

            Make your selection using SPACE key

            Get your results in a list called PnC.selectedX




    """


class point_n_click_plot():
    """
    This method allow to do "point and click" on a plot, to localize offsets 
    for instance
    
    Usage
    -----
    Data have to be ploted already in a figure
    
    .. code-block:: python
    
        PnC = point_n_click_plot()
        multi , cid = PnC(fig=1,Xdata_are_time=True)
        PnC.selectedX

    i.e.
    
    Create an object point_n_click_plot (here it is PnC in the exemple below) \n
    Call the object like a function with as 1st argument the id of the plot figure or the plot figure itself \n
    Make your selection using the SPACE key \n
    Get your results in a list called PnC.selectedX   
    
    Important
    ---------
    
    cursor objects (i.e. multi & cid)
    must be stored as global variables when you call the method
    like this :

    .. code-block:: python
    
        multi , cid = PnC(fig=1)
    
    """
    



    def __init__(self):
        self.selectedX = []
        self.ver_bar_stk = []


    def __call__(self,fig=1,Xdata_are_time=True):
        """
        IMPORTANT : cursor objects (i.e. multi & cid)
                    must be stored as global variables when you call the method
                    like this :
                    multi , cid = PnC(fig=1)
        """

        if type(fig) is int:
            figobj = plt.figure(fig)
        elif type(fig) is matplotlib.figure.Figure:
            figobj = fig

        log.info("press SPACE to record a X-value, \n       press R to Remove the previously recorded one")


        def onclick_discont(event):

            if event.key == ' ':
                if Xdata_are_time:
                    ix, iy = matplotlib.dates.num2date(event.xdata).replace(tzinfo=None),event.ydata
                else:
                    ix, iy = event.xdata , event.ydata

                log.info("X value recorded : " , ix)

                for ax in figobj.axes:
                    out_bar_list = stats.plot_vertical_bar_ax([ix],ax,"b",
                                                             linewidth=1)

                    self.ver_bar_stk.append(out_bar_list[0])

                self.selectedX.append(ix)

                plt.draw()

            elif event.key in ('r','R') and len(self.selectedX) > 0:
                last = self.selectedX[-1]

                self.selectedX.remove(last)

                last_bars = self.ver_bar_stk[-len(figobj.axes):]
                
                for bar in last_bars:
                    self.ver_bar_stk.remove(bar)    
                    bar.remove()
                    
                log.info("value removed : " , last )

                plt.draw()

            return None

        multi = MultiCursor(figobj.canvas, figobj.axes , color='k', lw=1)
        cid   = figobj.canvas.mpl_connect('key_press_event', onclick_discont)

        return multi , cid



# https://chat.deepseek.com/a/chat/s/4ac9a75c-2fe7-4a36-b206-a2c9c6fb6caa

import numpy as np
from netCDF4 import Dataset
import datetime as dt

def create_oceansites_netcdf(output_file):
    """
    Create an OceanSITES-compliant NetCDF file
    
    Parameters:
    output_file (str): Path to the output NetCDF file
    """
    
    # Create the NetCDF file
    rootgrp = Dataset(output_file, 'w', format='NETCDF4')
    
    now = dt.datetime.now(dt.UTC)
    
    # ======================
    # Global Attributes (Metadata)
    # ======================
    rootgrp.title = "OceanSITES data file"
    rootgrp.naming_authority = "OceanSITES"
    rootgrp.id = "SITE_CODE-PLATFORM_CODE-FILE_CODE"
    rootgrp.data_type = "OceanSITES time-series data"
    rootgrp.format_version = "1.4"
    rootgrp.Conventions = "CF-1.6, OceanSITES-1.3"
    rootgrp.date_created = now.strftime("%Y-%m-%dT%H:%M:%SZ")
    rootgrp.history = "Created " + now.strftime("%Y-%m-%d")
    
    # Data source information
    rootgrp.source = "mooring"
    rootgrp.institution = "Your Institution"
    rootgrp.institution_references = "https://www.your-institution.org"
    rootgrp.contact = "user@your-institution.org"
    rootgrp.references = "http://www.oceansites.org"
    
    # Data quality information
    rootgrp.QC_indicator = "mixed"
    rootgrp.QC_procedures = "OceanSITES QC manual v1.2"
    
    # Position information
    rootgrp.geospatial_lat_min = -30.5
    rootgrp.geospatial_lat_max = -30.5
    rootgrp.geospatial_lon_min = 15.2
    rootgrp.geospatial_lon_max = 15.2
    rootgrp.geospatial_vertical_min = 5.0
    rootgrp.geospatial_vertical_max = 500.0
    
    # Time information
    rootgrp.time_coverage_start = "2020-01-01T00:00:00Z"
    rootgrp.time_coverage_end = "2020-01-31T23:59:59Z"
    
    # ======================
    # Dimensions
    # ======================
    time_dim = rootgrp.createDimension('TIME', None)  # Unlimited dimension
    depth_dim = rootgrp.createDimension('DEPTH', 10)  # 10 depth levels
    
    # ======================
    # Coordinate Variables
    # ======================
    
    # TIME variable - note fill_value is set during creation
    time_var = rootgrp.createVariable('TIME', 'f8', ('TIME',), fill_value=-9999.0)
    time_var.standard_name = "time"
    time_var.long_name = "Time"
    time_var.units = "days since 1950-01-01 00:00:00 UTC"
    time_var.calendar = "gregorian"
    time_var.axis = "T"
    time_var.valid_min = 0.0
    time_var.valid_max = 90000.0
    
    # DEPTH variable - note fill_value is set during creation
    depth_var = rootgrp.createVariable('DEPTH', 'f4', ('DEPTH',), fill_value=-9999.0)
    depth_var.standard_name = "depth"
    depth_var.long_name = "Depth below sea level"
    depth_var.units = "meters"
    depth_var.axis = "Z"
    depth_var.positive = "down"
    depth_var.valid_min = 0.0
    depth_var.valid_max = 12000.0
    
    # LATITUDE variable (scalar) - note fill_value is set during creation
    lat_var = rootgrp.createVariable('LATITUDE', 'f4', fill_value=-9999.0)
    lat_var.standard_name = "latitude"
    lat_var.long_name = "Latitude of the measurement"
    lat_var.units = "degrees_north"
    lat_var.valid_min = -90.0
    lat_var.valid_max = 90.0
    
    # LONGITUDE variable (scalar) - note fill_value is set during creation
    lon_var = rootgrp.createVariable('LONGITUDE', 'f4', fill_value=-9999.0)
    lon_var.standard_name = "longitude"
    lon_var.long_name = "Longitude of the measurement"
    lon_var.units = "degrees_east"
    lon_var.valid_min = -180.0
    lon_var.valid_max = 180.0
    
    # ======================
    # Data Variables
    # ======================
    
    # Temperature - fill_value set during creation
    temp_var = rootgrp.createVariable('TEMP', 'f4', ('TIME', 'DEPTH'), fill_value=-9999.0)
    temp_var.standard_name = "sea_water_temperature"
    temp_var.long_name = "Sea temperature"
    temp_var.units = "degree_Celsius"
    temp_var.valid_min = -2.5
    temp_var.valid_max = 40.0
    temp_var.coordinates = "TIME DEPTH LATITUDE LONGITUDE"
    temp_var.QC_indicator = "1"
    temp_var.QC_procedure = "OceanSITES QC manual v1.2"
    
    # Salinity - fill_value set during creation
    psal_var = rootgrp.createVariable('PSAL', 'f4', ('TIME', 'DEPTH'), fill_value=-9999.0)
    psal_var.standard_name = "sea_water_practical_salinity"
    psal_var.long_name = "Practical salinity"
    psal_var.units = "1"
    psal_var.valid_min = 0.0
    psal_var.valid_max = 41.0
    psal_var.coordinates = "TIME DEPTH LATITUDE LONGITUDE"
    psal_var.QC_indicator = "1"
    psal_var.QC_procedure = "OceanSITES QC manual v1.2"
    
    # ======================
    # Fill with example data
    # ======================
    
    # Time data (30 days)
    time_data = np.arange(0, 30, 1)
    time_var[:] = time_data
    
    # Depth data (10 levels from 5 to 500m)
    depth_data = np.linspace(5, 500, 10)
    depth_var[:] = depth_data
    
    # Position data
    lat_var[:] = -30.5
    lon_var[:] = 15.2
    
    # Create temperature and salinity data with some random variation
    temp_data = 15 + 5 * np.random.randn(30, 10)
    psal_data = 35 + 0.5 * np.random.randn(30, 10)
    
    temp_var[:, :] = temp_data
    psal_var[:, :] = psal_data
    
    # ======================
    # Close the file
    # ======================
    rootgrp.close()
    print(f"OceanSITES NetCDF file created: {output_file}")

# Example usage
if __name__ == "__main__":
    create_oceansites_netcdf("/home/psakicki/aaa_FOURBI/out.nc")
    
    



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: psakic

This sub-module of geodezyx.files_rw contains reading functions to 
import files containing geodetic time series.

it can be imported directly with:
from geodezyx import files_rw

The GeodeZYX Toolbox is a software for simple but useful
functions for Geodesy and Geophysics under the GNU LGPL v3 License

Copyright (C) 2019 Pierre Sakic et al. (IPGP, sakic@ipgp.fr)
GitHub repository :
https://github.com/GeodeZYX/geodezyx-toolbox
"""

########## BEGIN IMPORT ##########
#### External modules
import datetime as dt
import glob
import gzip
#### Import the logger
import logging
import os
import re

import dateutil
import numpy as np
import pandas as pd
import scipy

#### geodeZYX modules
from geodezyx import conv
from geodezyx import files_rw
from geodezyx import reffram
from geodezyx import time_series
from geodezyx import utils

log = logging.getLogger('geodezyx')


##########  END IMPORT  ##########

def read_all_points(filein):
    """selectionne automatiquement le type de fichier brut en entrée
       INPUT  : chemin du fichier brut de POINTS
       OUTPUT : Une TimeSeriePoint """

    firstline = open(filein).readline()

    if re.compile('RTKLIB').search(firstline):
        tsout = read_rtklib(filein)

    elif re.compile('Kinematic Processing').search(firstline):
        tsout = read_gipsy_apps(filein)

    elif re.compile('tdp.llh').search(filein):
        tsout = read_gipsy_bosser(filein)

    elif re.compile('STA').search(firstline) or re.compile('tdp').search(filein) or re.compile('TRPAZ').search(
            firstline):
        tsout = read_gipsy_tdp(filein)

    elif re.compile('YY  MM DD HR MIN').search(firstline):
        tsout = read_track(filein)

    elif re.compile('Latitude').search(firstline):
        tsout = read_sonardyne_posi(filein)

    elif re.compile('Heading').search(firstline):
        tsout = read_sonardyne_attitude(filein)

    elif re.compile(r'\*\*\* warning').search(firstline):
        tsout = read_gins(filein, 'kine')

    elif re.compile('#GINS_VERSION').search(firstline):
        tsout = read_gins_solution(filein)

    elif re.compile('OCTANS_ATTITUDE').search(firstline):
        tsout = read_qinsy(filein, 2014, 0o4, 0o4)

    elif re.compile('PBO Station Position Time Series').search(firstline):
        tsout = read_pbo_pos(filein)

    elif re.compile('latitude_degre_decimal').search(firstline):
        tsout = read_nrcan_csv(filein)

    elif re.compile('--------------------------------------------------').search(firstline):  # NDLR : Best parser ever
        tsout = read_nrcan_pos(filein)

    else:
        log.error("pas de motif valide pour lect. auto")
        log.info(filein)
        log.info(firstline)

    return tsout


def read_all_obs(filein):
    """selectionne automatiquement le type de fichier brut en entrée
       INPUT  : chemin du fichier brut de observations génériques
       OUTPUT : Une LISTE de TimeSerieObs """

    firstline = open(filein).readline()

    if re.compile('Heading').search(firstline):
        tsout = read_sonardyne_attitude(filein)

    else:
        log.error("pas de motif valide pour lect. auto")

    return tsout


def read_rtklib(filein):
    """lit un fichier de type RTKLIB
       INPUT  : chemin du fichier brut
       OUTPUT : Une TimeSeriePoint """

    tsout = time_series.TimeSeriePoint()

    for line in open(filein):

        if re.compile('e-baseline').search(line):
            initype = 'ENU'
        elif re.compile('x-ecef').search(line):
            initype = 'XYZ'
        elif 'latitude(deg)' in line:
            initype = 'FLH'

        if re.compile('%  UTC').search(line):
            dUTCGPS = 16
            log.warning('MAYBE A WRONG LEAP SECOND !!!!')
        elif re.compile('%  GPST').search(line):
            dUTCGPS = 0

        if line[0] == '%':
            continue

        fields = line.split()
        date1 = re.findall(r"[\w']+", fields[0] + ':' + fields[1])
        date2 = tuple([int(d) for d in date1[:-1]] + [int(date1[-1][:6])])

        T = (dt.datetime(date2[0], date2[1], date2[2], date2[3], date2[4], date2[5], date2[6]) + dt.timedelta(
            seconds=dUTCGPS))
        A = (float(fields[2]))
        B = (float(fields[3]))
        C = (float(fields[4]))
        if initype == 'XYZ':
            sA = (float(fields[7]))
            sB = (float(fields[8]))
            sC = (float(fields[9]))
        elif initype == 'FLH':
            sA = (float(fields[7]))
            sB = (float(fields[8]))
            sC = (float(fields[9]))
            sA, sB, sC = conv.sigma_enu2geo(A, B, C, sA, sB, sC)

        point = time_series.Point(A, B, C, T, initype, sA, sB, sC)
        point.anex['sdAB'] = float(fields[10])
        point.anex['sdBC'] = float(fields[11])
        point.anex['sdAC'] = float(fields[12])

        tsout.add_point(point)

    tsout.meta_set(filein)

    if initype == 'ENU':
        tsout.boolENU = True

    try:
        inpfilis = utils.grep(filein, 'inp file')
        tsout.anex['rover'] = os.path.basename(inpfilis[0].split()[-1])[0:4].upper()
        tsout.anex['base'] = os.path.basename(inpfilis[1].split()[-1])[0:4].upper()
    except:
        pass

    return tsout


#       _ _____  _          _______ _____ _____   _______     __  ______ _ _
#      | |  __ \| |        / / ____|_   _|  __ \ / ____\ \   / / |  ____(_) |
#      | | |__) | |       / / |  __  | | | |__) | (___  \ \_/ /  | |__   _| | ___  ___
#  _   | |  ___/| |      / /| | |_ | | | |  ___/ \___ \  \   /   |  __| | | |/ _ \/ __|
# | |__| | |    | |____ / / | |__| |_| |_| |     ____) |  | |    | |    | | |  __/\__ \
#  \____/|_|    |______/_/   \_____|_____|_|    |_____/   |_|    |_|    |_|_|\___||___/


def read_gipsy_tdp(filein):
    """
    Read legacy Gipsy TDP (Time Dependent Parameter) File

    Parameters
    ----------
    filein : str
        input file path.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.
    """

    X, Y, Z = np.nan, np.nan, np.nan
    Tx, Ty, Tz, T = np.nan, np.nan, np.nan, np.nan
    sX, sY, sZ = np.nan, np.nan, np.nan

    tsout = time_series.TimeSeriePoint()

    for line in open(filein):

        fields = line.split()

        if fields[4] == 'STA' and fields[5] == 'Z':
            Tz = conv.tgipsy2dt(fields[0])
            Z = (float(fields[2]) * 1000)
            sZ = (float(fields[3]) * 1000)

        if fields[4] == 'STA' and fields[5] == 'Y':
            Ty = conv.tgipsy2dt(fields[0])
            Y = (float(fields[2]) * 1000)
            sY = (float(fields[3]) * 1000)

        if fields[4] == 'STA' and fields[5] == 'X':
            Tx = conv.tgipsy2dt(fields[0])
            X = (float(fields[2]) * 1000)
            sX = (float(fields[3]) * 1000)
            STAT = fields[6]

        if Tx == Ty == Tz:
            T = Tx
            point = time_series.Point(X, Y, Z, T, 'XYZ', sX, sY, sZ)
            tsout.add_point(point)

            Tx = 111
            Ty = 222
            Tz = 333

    tsout.meta_set(filein, stat=STAT)

    return tsout


def read_gipsy_tdp_list(filelistin):
    """
    Read Several GIPSY TDP (Time Dependent Parameter) Files
    

    Parameters
    ----------
    filelistin : list
        input file paths in a list.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.    
    """

    tslist = []
    for fil in filelistin:
        ts = read_gipsy_tdp(fil)
        tslist.append(ts)

    tsout = time_series.merge_ts(tslist)

    stat = list(set([ts.stat for ts in tslist]))[0]

    tsout.meta_set("", stat)

    return tsout


def read_gipsyx_tdp(filein):
    """
    Read GipsyX TDP (Time Dependent Parameter) File

    Parameters
    ----------
    filein : str
        input file path.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.
    """

    X, Y, Z = np.nan, np.nan, np.nan
    tx, ty, tz, T = np.nan, np.nan, np.nan, np.nan
    s_x, s_y, s_z = np.nan, np.nan, np.nan

    tsout = time_series.TimeSeriePoint()

    try:
        fil = open(filein)
    except Exception as e:
        log.error("unable to open %s", filein)
        raise e

    for line in fil:

        fields = line.split()

        if len(fields) == 0:
            continue

        attribs = fields[-1].split(".")

        if attribs[1] == 'Station' and attribs[5] == 'Z':
            tz = conv.tgipsy2dt(fields[0])
            Z = (float(fields[2]))
            s_z = (float(fields[3]))

        if attribs[1] == 'Station' and attribs[5] == 'Y':
            ty = conv.tgipsy2dt(fields[0])
            Y = (float(fields[2]))
            s_y = (float(fields[3]))

        if attribs[1] == 'Station' and attribs[5] == 'X':
            tx = conv.tgipsy2dt(fields[0])
            X = (float(fields[2]))
            s_x = (float(fields[3]))
            stat = attribs[2]

        if tx == ty == tz:
            T = tx
            point = time_series.Point(X, Y, Z, T, 'XYZ', s_x, s_y, s_z)
            tsout.add_point(point)

            tx = np.nan
            ty = np.nan
            tz = np.nan

    tsout.meta_set(filein, stat=stat)
    tsout.sort()

    return tsout


def read_gipsyx_tdp_list(filelistin):
    """
    Read Several GIPSYX TDP (Time Dependent Parameter) Files
    

    Parameters
    ----------
    filelistin : list
        input file paths in a list.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.    
    """

    tslist = []
    for fil in filelistin:
        ts = read_gipsyx_tdp(fil)
        tslist.append(ts)

    tsout = time_series.merge_ts(tslist)

    stat = list(set([ts.stat for ts in tslist]))[0]

    tsout.meta_set("", stat)

    return tsout


def read_gipsy_gdcov(filein):
    X, Y, Z = np.nan, np.nan, np.nan
    Tx, Ty, Tz, T = np.nan, np.nan, np.nan, np.nan
    sX, sY, sZ = np.nan, np.nan, np.nan

    tsout = time_series.TimeSeriePoint()

    F = open(filein)
    L = F.readlines()

    ### parameters search
    param = int(L[0].split()[0])

    Lparam = L[1:param + 1]
    Lcovar = L[param + 2:]

    for line in Lparam:
        fields = line.split()
        attribs = fields[1].split(".")

        if attribs[1] == 'STA' and attribs[-1] == 'Z':
            Tz = conv.tgipsy2dt(fields[2])
            Z = (float(fields[3]))
            sZ = (float(fields[4]))

        if attribs[1] == 'STA' and attribs[-1] == 'Y':
            Ty = conv.tgipsy2dt(fields[2])
            Y = (float(fields[3]))
            sY = (float(fields[4]))

        if attribs[1] == 'STA' and attribs[-1] == 'X':
            Tx = conv.tgipsy2dt(fields[2])
            X = (float(fields[3]))
            sX = (float(fields[4]))
            STAT = attribs[0]

        if Tx == Ty == Tz:
            T = Tx
            point = time_series.Point(X, Y, Z, T, 'XYZ', sX, sY, sZ)
            tsout.add_point(point)

            Tx = np.nan
            Ty = np.nan
            Tz = np.nan

    tsout.meta_set(filein, stat=STAT)

    return tsout


def read_gipsy_gdcov_list(filelistin):
    tslist = []
    for fil in filelistin:
        ts = read_gipsy_gdcov(fil)
        tslist.append(ts)

    tsout = time_series.merge_ts(tslist)

    stat = list(set([ts.stat for ts in tslist]))[0]

    tsout.meta_set("", stat)

    return tsout


def read_gipsyx_xfile(filein):
    """
    Read GIPSYX X file i.e. the transformation parameters and their 
    residuals
    

    Parameters
    ----------
    filein : str
        input file path.
        Can handle gz compressed files
        
    Returns
    -------
    df_trans_out : DataFrame
        Helmert transformation parameters and their sigmas.
    df_resid_out : DataFrame
        Coordinates residuals (not implemented yet).

    """

    fname = os.path.basename(filein)

    date = conv.date_string_2_dt(fname[:11])

    if filein[-2:] in ("gz", "GZ"):
        F = gzip.open(filein, "r+")
        lines = [e.decode('utf-8') for e in F]
    else:
        F = open(filein, "r+")
        lines = F.readlines()

    df_trans_out = pd.DataFrame()
    df_resid_out = pd.DataFrame()

    df_trans_out.loc[0, "epoch"] = date

    for l in lines:
        #### get transform parameters
        if re.search(' = ', l):
            l2 = l.split()
            label = l2[0]
            val = float(l2[2])
            df_trans_out.loc[0, label] = val

            if len(l2) > 3:
                val_sigma = float(l2[4])
                df_trans_out.loc[0, "s" + label] = val_sigma

        #### get residual values
        # l_resid = []
        # if re.search('^ ( POS| RES)', l):
        #     l_resid.append(l)

        # df_resid_out = pd.read_csv(StringIO("\n".join(l_resid[:-1])))

    return df_trans_out, df_resid_out


def read_gipsyx_xfile_list(filelistin):
    """
    Read several GIPSYX X files i.e. the transformation parameters and their 
    residuals

    Parameters
    ----------
    filelistin : list
        input file paths in a list.
        Can handle gz compressed files


    Returns
    -------
    df_trans_out : DataFrame
        Helmert transformation parameters and their sigmas.
    df_resid_out : DataFrame
        Coordinates residuals (not implemented yet).

    """
    dflist = []
    for fil in filelistin:
        df_trans_mono, df_resid_mono = read_gipsyx_xfile(fil)
        dflist.append(df_trans_mono)

    df_trans_out = pd.concat(dflist)
    df_trans_out.reset_index(drop=True, inplace=True)
    df_resid_out = pd.DataFrame()

    return df_trans_out, df_resid_out


def read_gipsy_bosser(filein):
    """
    Read p. Bosser (@ENSTA Brest) File (GIPSY)

    Parameters
    ----------
    filein : str
        input file path.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.
    """

    F, L, H = 0, 0, 0
    T = 0
    sF, sL, sH = 0, 0, 0

    tsout = time_series.TimeSeriePoint()

    for line in open(filein):
        f = line.split()

        y = int(f[0])
        doydec = float(f[1])
        doy = int(doydec)

        T = conv.doy2dt(y, doy) + dt.timedelta(days=(doydec - doy))

        F = np.rad2deg(float(f[2]))
        L = np.rad2deg(float(f[3]))
        H = float(f[4])
        RMS = float(f[8])

        point = time_series.Point(F, L, H, T, 'FLH')
        point.anex['RMS'] = RMS
        tsout.add_point(point)

    tsout.meta_set(filein, stat='STAT')

    return tsout


def read_gipsy_apps(filein):
    """
    Read GIPSY APPS (Online tool) File

    Parameters
    ----------
    filein : str
        input file path.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.
    """

    tsout = time_series.TimeSeriePoint()

    for l in open(filein):
        if l[0] == '#' or ('Kinematic Processing' in l):
            continue
        f = l.split()

        date_lis = [int(float(e)) for e in f[1].split(':')]

        T = dt.datetime(*date_lis)

        X = float(f[2])
        sX = float(f[3])
        Y = float(f[4])
        sY = float(f[5])
        Z = float(f[6])
        sZ = float(f[7])

        point = time_series.Point(X, Y, Z, T, 'XYZ', sX, sY, sZ)
        tsout.add_point(point)

    tsout.meta_set(filein)
    return tsout


def read_jpl_timeseries_solo(latlonrad_files_list):
    tsout = time_series.TimeSeriePoint()

    latpath = [f for f in latlonrad_files_list if '.lat' in f][0]
    lonpath = [f for f in latlonrad_files_list if '.lon' in f][0]
    radpath = [f for f in latlonrad_files_list if '.rad' in f][0]

    latfile = open(latpath)
    lonfile = open(lonpath)
    radfile = open(radpath)

    for llat, llon, lrad in zip(latfile, lonfile, radfile):
        flat = [float(e) for e in llat.split()[0:3]]
        flon = [float(e) for e in llon.split()[0:3]]
        frad = [float(e) for e in lrad.split()[0:3]]

        if not (flat[0] == flon[0] == frad[0]):
            log.error("%s %s %s", flat[0], flon[0], frad[0])
            log.error('Time dont corresponds !!!')
            raise Exception

        statlat = os.path.basename(latpath).split('.')[0]
        statlon = os.path.basename(lonpath).split('.')[0]
        statrad = os.path.basename(radpath).split('.')[0]

        if not (statlat == statlon == statrad):
            log.info("%s %s %s", statlat, statlon, statrad)
            log.error('Station name do not corresponds !!!')
            raise Exception

        T = conv.year_decimal2dt(flat[0])

        N = flat[1] * 10 ** -2
        E = flon[1] * 10 ** -2
        U = frad[1] * 10 ** -2

        sN = flat[2] * 10 ** -2
        sE = flon[2] * 10 ** -2
        sU = frad[2] * 10 ** -2

        point = time_series.Point(E, N, U, T, 'ENU', sE, sN, sU)

        tsout.boolENU = True
        tsout.add_point(point)

    tsout.stat = statlat

    return tsout


#  __  __ _____ _______  _______          __  __ _____ _______   ______ _ _
# |  \/  |_   _|__   __|/ / ____|   /\   |  \/  |_   _|__   __| |  ____(_) |
# | \  / | | |    | |  / / |  __   /  \  | \  / | | |    | |    | |__   _| | ___  ___
# | |\/| | | |    | | / /| | |_ | / /\ \ | |\/| | | |    | |    |  __| | | |/ _ \/ __|
# | |  | |_| |_   | |/ / | |__| |/ ____ \| |  | |_| |_   | |    | |    | | |  __/\__ \
# |_|  |_|_____|  |_/_/   \_____/_/    \_\_|  |_|_____|  |_|    |_|    |_|_|\___||___/


def read_track_2(filein, site_name=None):
    """
    Read a kinematic track file

    Parameters
    ----------
    filein : str
        path of the file.

    Returns
    -------
    DF : Pandas DataFrame

    """
    DF = pd.read_csv(filein, delim_whitespace=True, skiprows=[1, 2])
    DF.columns = ['year', 'month', 'day', 'hour', 'minute', 'second',
                  'dX', 'dX_std', 'dY', 'dY_std', 'dZ', 'dZ_std',
                  'rms', 'dd', 'atm', 'atm_std', 'fract_doy',
                  'n_epoch', 'BF', 'not', 'f', 'rho_ua', 'null']

    DF.drop("null", axis=1, inplace=True)

    Epoch = conv.ymdhms2dt(DF.year, DF.month, DF.day,
                           DF.hour, DF.minute, DF.second)
    DF['epoch'] = Epoch

    if site_name:
        DF['site'] = site_name()

    return DF


def read_track(filein):
    """
    Read GAMIT/TRACK File

    Parameters
    ----------
    filein : str
        input file path.

    Returns
    -------
    tsout : TimeSeries Object
        output TimeSerie.
    """

    tsout = time_series.TimeSeriePoint()

    for line in open(filein):

        if re.compile('dNorth').search(line):
            initype = 'ENU'
        elif re.compile('dX').search(line):
            initype = 'XYZ'

        if line[0] != ' ':
            continue

        fields = line.split()

        T = (dt.datetime(int(fields[0]),
                         int(fields[1]),
                         int(fields[2]),
                         int(fields[3]),
                         int(fields[4]),
                         int(fields[5].split('.')[0]),
                         int(np.round(float(fields[5].split('.')[1]), 4))))
        A = (float(fields[6]))
        B = (float(fields[8]))
        C = (float(fields[10]))

        sA = (float(fields[7]))
        sB = (float(fields[9]))
        sC = (float(fields[11]))

        # On inverse A et B car NEU => ENU
        if initype == 'ENU':
            point = time_series.Point(B, A, C, T, initype, sB, sA, sC)
        elif initype == 'XYZ':
            point = time_series.Point(A, B, C, T, initype, sA, sB, sC)
        else:
            "ERR : track_read : bad initype"

        tsout.add_point(point)

    if initype == 'ENU':
        tsout.boolENU = True

    tsout.meta_set(filein)

    tsout.anex['rover'] = tsout.name.split('.')[-2].upper()

    # recherche de la base
    try:
        if '.LC' in filein:
            stat_n_base_lc_files = glob.glob(filein[:-8] + '*')
            stat_n_base_lc_files.remove(filein)
            statbase = os.path.basename(stat_n_base_lc_files[0])[-7:-3]
            tsout.anex['base'] = statbase
        elif '.L1+L2' in filein:
            stat_n_base_l1l2_files = glob.glob(filein[:-11] + '*')
            stat_n_base_l1l2_files.remove(filein)
            statbase = os.path.basename(stat_n_base_l1l2_files[0])[-10:-6]
            tsout.anex['base'] = statbase
    except:
        log.warning('unable to find the base for the TRACK experience')
        log.info(filein)
        pass

    return tsout


#   _____ _   _ ______  _____    _______ _____ _   _  _____   ______ _ _
#  / ____| \ | |  ____|/ ____|  / / ____|_   _| \ | |/ ____| |  ____(_) |
# | |    |  \| | |__  | (___   / / |  __  | | |  \| | (___   | |__   _| | ___  ___
# | |    | . ` |  __|  \___ \ / /| | |_ | | | | . ` |\___ \  |  __| | | |/ _ \/ __|
# | |____| |\  | |____ ____) / / | |__| |_| |_| |\  |____) | | |    | | |  __/\__ \
#  \_____|_| \_|______|_____/_/   \_____|_____|_| \_|_____/  |_|    |_|_|\___||___/

def read_gins_solution(filein, mode="cinematic"):
    """
    Read a GINS solution file

    Parameters
    ----------
    filein : str
        path of the input file.
    mode : str, optional
        cinematic : retrun a TimeSerie
        static : retrun a point
        
    Returns
    -------
    TimeSeries or Point object
        output Time Series.

    """

    F = open(filein)

    pts_list_tmp = []
    namestat = "XXXX"

    datexere = re.search(r"[0-9]{6}_[0-9]{6}", os.path.basename(filein))
    if not datexere:
        datexe = '991231_235959'
    else:
        datexe = datexere[0]

    ginsvers = 'unknown'

    for l in F:
        f = l.split()

        if 'STATION_NAME' in l:
            # get per default 4char name
            namestat = f[1] if len(f) > 1 else f[-1]

            # try to catch the 9char name in the filename 
            filnam = os.path.basename(filein)
            reout = re.search(namestat + "[0-9]{2}[A-Z]{3}", filnam)
            if reout:
                namestat = reout.group(0)

        if "GINS_VERSION" in l:
            ginsvers = f[1]
        elif l[0] == '#':
            continue

        # Traw = float(f[2])

        if 'XYZ_SOL' in l:
            coordstype = 'XYZ'
            X = float(f[4])
            Y = float(f[6])
            Z = float(f[8])

            sX = np.sqrt(float(f[9]))
            sY = np.sqrt(float(f[10]))
            sZ = np.sqrt(float(f[11]))

            if "T24:00:00.000" in f[1]:  # Manage the special case if we are at the border bw 2 days
                Txyz = conv.string_date2dt(f[1][:10]) + dt.timedelta(days=+1) + dt.timedelta(seconds=-19)
            else:
                Txyz = conv.string_date2dt(f[1]) + dt.timedelta(seconds=-19)

            point = time_series.Point(X, Y, Z, Txyz, coordstype, sX, sY, sZ, name=namestat)

            point.anex['sdXY'] = float(f[12])
            point.anex['sdXZ'] = float(f[13])
            point.anex['sdYZ'] = float(f[14])
            point.anex['sol_path'] = filein
            point.anex['dateofexe'] = datexe
            point.anex['gins_version'] = ginsvers


        elif 'FLH_SOL' in l:
            coordstype = 'FLH'
            F = float(f[4])
            L = float(f[6])
            H = float(f[8])

            sF = np.rad2deg(np.sqrt(float(f[9])))
            sL = np.rad2deg(np.sqrt(float(f[10])))
            sH = np.sqrt(float(f[11]))

            point.F, point.L, point.H = F, L, H
            point.sF, point.sL, point.sH = sF, sL, sH

            # point.X , point.Y , point.Z = X,Y,Z

            point.anex['sdFL'] = float(f[12])
            point.anex['sdFH'] = float(f[13])
            point.anex['sdLH'] = float(f[14])

            pts_list_tmp.append(point)

    #### End of reading, export
    if not pts_list_tmp:
        log.warning("no point found in: %s. Returns None", filein)
        return None

    if mode == "cinematic":
        tsout = time_series.TimeSeriePoint()
        for point in pts_list_tmp:
            tsout.add_point(point)
        tsout.meta_set(filein, namestat)
        return tsout

    elif mode == "static":
        pt_out = pts_list_tmp[0]
        return pt_out


def read_gins_solution_multi(filein_list, return_dict=True):
    """
    Read multiple GINS solution files and return a dictionary or list of time series.

    Parameters
    ----------
    filein_list : list
        List of input file paths.
    return_dict : bool, optional
        If True, return a dictionary of time series.
        If False, return a list of time series. Defaults to True.

    Returns
    -------
    dict or list
        A dictionary or list of TimeSeriePoint objects,
        depending on the value of return_dict.
    """

    filein_list = sorted(filein_list)
    points_list = []
    statname_stk = []

    for fil in filein_list:
        pt_day = read_gins_solution(fil, mode="static")
        if not pt_day:
            continue
        points_list.append(pt_day)
        statname_stk.append(pt_day.name)

    statname_uniq = sorted(list(set(statname_stk)))
    ts_dict = dict()

    for point in points_list:
        if not point.name in ts_dict.keys():
            ts_dict[point.name] = time_series.TimeSeriePoint(stat=point.name)
            ts_dict[point.name].meta_set(stat=point.name, name=point.name)
        ts_dict[point.name].add_point(point)

    ## return a dictionnary of TS or a list of TS
    if return_dict:
        return ts_dict
    else:
        ts_list = []
        for k, val in ts_dict.items():
            ts_list.append(val)
        return ts_list


def read_gins(filein, kineorstatic='kine', flh_in_rad=True,
              force_get_convergence=False, kf_result=False):
    '''
    Static : donne un point
    Kinematic : donne une TS

    force_get_convergence : if there is a bug about
    'COORDONNEES DES STATIONS AJUSTEES EN HAUTE FREQUENCE' field, it is in the listing
    but empty ... so we force the retreive of the 'c o n v e r g e n c e' part
    '''

    if '.prepars' in filein:
        log.warning('%s seems to be a prepars file, are you sure of what you are doing ?', filein)

    # Pour le static, il y a des blancs en fin de ligne ...

    if kineorstatic == 'kine':
        # regex = '\[S[PLHXYZ] .*\]$'
        regex = r'\[S[PLHXYZ] .*\]$'
        tsout = time_series.TimeSeriePoint()
        if kf_result:
            regex = r'\[S[PLHXYZ][E ].*\]     $'
    elif kineorstatic == 'static':
        # regex = '\[S[PLHXYZ] .*\]     $'
        regex = r'\[S[PLHXYZ][E ].*\]     $'
        tsout = time_series.TimeSeriePoint()
    else:
        log.error("ERR")

    A, B, C = 0, 0, 0
    Ta, Tb, Tc, T = 111, 222, 333, 0
    sA, sB, sC = 0, 0, 0

    if kf_result:
        regex = r'\[S[PLHXYZ][E ].*\]     $'

    # Specific si 2ble convergence
    grep_conv = utils.grep(filein, 'c o n v e r g e n c e')
    if len(grep_conv) == 2:
        IPPmode = True
        converg_compt = 0
        log.info('%s have 2  c o n v e r g e n c e  fields', os.path.basename(filein))
        log.info("keeping the last one")
    else:
        IPPmode = False

    # Specific si Ajustement Final
    greped_adj = utils.grep(filein, 'COORDONNEES DES STATIONS AJUSTEES EN HAUTE FREQUENCE')
    if force_get_convergence:
        FinalAdj_mode = False
    elif len(greped_adj) != 0:
        FinalAdj_mode = True
        FinalAdj_found = False
        log.info('%s have a COORD STAT AJ EN HTE FREQ  field', os.path.basename(filein))
        log.info("     keeping this one (and not the convergence)")
    else:
        FinalAdj_mode = False

    fileopened = open(filein, encoding="ISO-8859-1")

    regex_valid_line_count = 0

    for line in fileopened:

        if re.compile('__Nom__').search(line):
            nextline = next(fileopened).split()
            namestat = nextline[3]
            Xref = float(nextline[4])
            Yref = float(nextline[5])
            Zref = float(nextline[6])

            Fref, Lref, Href = conv.xyz2geo(Xref, Yref, Zref)

        #        if 'angles en deg' in line:
        #            flh_in_rad = False

        # Specific search
        if IPPmode and re.compile('c o n v e r g e n c e').search(line):
            converg_compt = converg_compt + 1
        if FinalAdj_mode and re.compile('COORDONNEES DES STATIONS AJUSTEES EN HAUTE FREQUENCE').search(line):
            FinalAdj_found = True

        # Specific skip
        if FinalAdj_mode and not FinalAdj_found:
            continue
        if IPPmode and converg_compt != 2:
            continue

        if "real" in line:
            rawexectime = line.split()[-1]

        if re.compile(regex).search(line):
            regex_valid_line_count += 1
            fields = line.split()

            if re.compile('[XYZ]').search(line):
                initype = 'XYZ'
                Aref, Bref, Cref = Xref, Yref, Zref
            elif re.compile('[PLH]').search(line):
                initype = 'FLH'
                Aref, Bref, Cref = Fref, Lref, Href
            else:
                log.error("wrong initype")

            if (float(fields[2]) == 0):
                continue

            if (fields[0] == 'stations'):
                pass

            # securité pour les lignes du type
            #  ------------------------------------------------------------------------------------------------------
            # stations      175  -0.574353783560234E+07  +/-   0.000000000000000E+00   1  [SX  1212001892701M005071]

            jour = int(line[108:110])
            h = int(line[110:112])
            m = int(line[112:114])
            s = int(line[114:116])
            yy = int(line[125:127])

            # gestion des années
            if 80 < yy <= 99:
                yy = yy + 1900
            else:
                yy = yy + 2000

            # pour le mois, si > sept (9), alors lettre ...
            mm = line[127]

            if mm == 'O':
                mm = 10
            elif mm == 'N':
                mm = 11
            elif mm == 'D':
                mm = 12
            else:
                mm = int(mm)

            try:
                if h == 24:
                    # cas exceptionnel ou on doit gerer minuit
                    # on retranche l'heure dans l'int en input et on l'ajoute dans le dt
                    Ttemp = (dt.datetime(yy, mm, jour, h - 1, m, s) + dt.timedelta(seconds=-19) + dt.timedelta(hours=1))
                else:
                    Ttemp = (dt.datetime(yy, mm, jour, h, m, s) + dt.timedelta(seconds=-19))

                if line[105] == 'X' or line[105] == 'p':
                    Ta = Ttemp
                    A = (float(fields[3]))
                    sA = (float(fields[4]))

                if line[105] == 'Y' or line[105] == 'L':
                    Tb = Ttemp
                    B = (float(fields[3]))
                    sB = (float(fields[4]))

                if line[105] == 'Z' or line[105] == 'H':
                    Tc = Ttemp
                    C = (float(fields[3]))
                    sC = (float(fields[4]))
            except ValueError as err:
                log.error('yy,mm,jour,h,m,s, %s %s %s %s %S ', yy, mm, jour, h, m, s)
                raise err

            if Ta == Tb == Tc:
                T = Ta
                if initype == 'FLH' and not FinalAdj_mode:
                    if flh_in_rad:
                        A = np.rad2deg(A)
                        B = np.rad2deg(B)
                        sA = np.rad2deg(sA)
                        sB = np.rad2deg(sB)
                if kf_result and 0:
                    A = A + Aref
                    B = B + Bref
                    C = C + Cref
                point = time_series.Point(A, B, C, T, initype, sA, sB, sC, name=namestat)

                Ta = 111
                Tb = 222
                Tc = 333

                if kineorstatic == 'static':
                    return point
                elif kineorstatic == 'kine':
                    tsout.add_point(point)
                else:
                    log.error("ERROR")

    if regex_valid_line_count == 0:
        log.warning("no valid line (with regex check) was found !!!")

    tsout.anex['exec_time'] = rawexectime
    tsout.meta_set(filein, namestat)
    return tsout


def gins_read_time(line):
    jour = int(line[108:110])
    h = int(line[110:112])
    m = int(line[112:114])
    s = int(line[114:116])
    yy = int(line[125:127])

    # gestion des années
    if 80 < yy <= 99:
        yy = yy + 1900
    else:
        yy = yy + 2000

    # pour le mois, si > sept (9), alors lettre ...
    mm = line[127]

    if mm == 'O':
        mm = 10
    elif mm == 'N':
        mm = 11
    elif mm == 'D':
        mm = 12
    else:
        mm = int(mm)

    try:
        if h == 24:
            # cas exceptionnel ou on doit gerer minuit
            # on retranche l'heure dans l'int en input et on l'ajoute dans le dt
            Ttemp = (dt.datetime(yy, mm, jour, h - 1, m, s) +
                     dt.timedelta(seconds=-19) + dt.timedelta(hours=1))
        else:
            Ttemp = (dt.datetime(yy, mm, jour, h, m, s) + dt.timedelta(seconds=-19))
    except:
        Ttemp = dt.datetime(1970, 1, 1)

    T = Ttemp

    return T


def gins_read_MZB(filein, return_df=False):
    """
    Read Mean Zeintal Bias in a GINS' listing
    """

    F = open(filein)

    regex = r'\[MZB.*\]     $'

    Tstk = []
    MZBstk = []
    sMZBstk = []
    NameStat = []

    for line in F:
        if re.compile('__Nom__').search(line):
            nextline = next(F).split()
            namestat = nextline[3]
            Xref = float(nextline[4])
            Yref = float(nextline[5])
            Zref = float(nextline[6])

        if re.compile(regex).search(line):
            fields = line.split()

            # [MZB  801980015051301 GPS]

            Traw = fields[6][-8:]
            yy = int(Traw[0:2])
            mm = int(Traw[2:4])
            dd = int(Traw[4:6])
            hh = int(Traw[6:8])

            if 80 < yy <= 99:
                yy = yy + 1900
            else:
                yy = yy + 2000

            T = dt.datetime(yy, mm, dd, hh)

            MZB = float(fields[3])
            sMZB = float(fields[4])

            Tstk.append(T)
            MZBstk.append(MZB)
            sMZBstk.append(sMZB)
            NameStat.append(namestat)

    if not return_df:
        return Tstk, MZBstk, sMZBstk, NameStat
    else:
        DF = pd.DataFrame((Tstk, MZBstk, sMZBstk, NameStat))
        DF = DF.T
        DF.columns = ('epoch', 'mzb', 'mzb_std', 'site')
        DF.mzb = DF.mzb.astype(float)
        DF.mzb_std = DF.mzb_std.astype(float)
        return DF


def gins_readTROPOZ(filein):
    """
    Read TROPOZ in a GINS' listing
    """

    L = utils.grep(filein, "TROPOZ COR_ZEN_ESTIM")
    DF = pd.DataFrame([e.split()[2:] for e in L]).astype(float)
    DF.columns = ['jjul_cnes', 'tropoz_std', 'tropoz']
    DF["epoch"] = conv.jjul_cnes2dt(DF['jjul_cnes']).dt.round('1s') - dt.timedelta(seconds=19)

    return DF


def write_ATM_GAMIT(Tstk, MZBstk, sMZBstk,
                    namestat, file_out):
    Fout = open(file_out, 'w+')
    for T, mzb, smzb in zip(Tstk, MZBstk, sMZBstk):
        yy = T.year
        mm = T.month
        dd = T.day
        hh = T.hour
        Line = 'ATM_ZEN X {}  1 {:4} {:2} {:2} {:2}  0  {:6.4f} +-   {:6.4f}    {:6.4f}\n'.format(namestat.upper(), yy,
                                                                                                  mm, dd, hh, mzb, smzb,
                                                                                                  mzb)
        Fout.write(Line)
    Fout.close()
    return file_out


def MZB_GINS_2_ATM_GAMIT(listing_in, path_out):
    Tstk, MZBstk, sMZBstk, namestat = gins_read_MZB(listing_in)
    doy, yy = conv.dt2doy_year(Tstk[0], str)
    file_out = os.path.join(path_out, '_'.join(('MZB_GINS_2_ATM_GAMIT', namestat, doy, yy, '.txt')))
    write_ATM_GAMIT(Tstk, MZBstk, sMZBstk,
                    namestat, file_out)

    return file_out


def read_gins_wrapper(input_list_or_path, flh_in_rad=True):
    """
    pour une liste de paths renvoie une liste de timeseries
    AUTANT DE TIMESERIES QUE DE LISTINGS

    INPUT :
    Une liste de fichiers
    Un path compatible avec glob

    les fichiers sans extensions .gins
    sont automatiquement exclus
    """

    if type(input_list_or_path) is str:
        gins_listings_list = glob.glob(input_list_or_path)
    else:
        gins_listings_list = input_list_or_path

    tslis = []
    for f in gins_listings_list:
        if not '.gins' in f:
            log.warning('WARN : no .gins ext, skipping')
            continue

        ts = read_gins(f, flh_in_rad=flh_in_rad)
        tslis.append(ts)

    return tslis


def convert_sp3_clk_2_GINS_clk(sp3_path_in,
                               clk_gins_out,
                               interpo_30sec=True,
                               return_as_DF=True):
    DF = files_rw.read_sp3(sp3_path_in)

    Fout = open(clk_gins_out, "w+")

    def write_GINS_signaletic_elt_clk(dt_in, sv,
                                      signaletic_name="MNG",
                                      add_19sec_to_dt_in=True):
        """
        very beta, only for GPS clk
        """

        if add_19sec_to_dt_in:
            dt_work = dt_in + dt.timedelta(seconds=19)
        else:
            dt_work = dt_in

        jjul = conv.dt2jjul_cnes(dt_work)
        sec_in_day = (dt_work - conv.jjul_cnes2dt(jjul)).seconds

        # MNG0000000jjjjjcccccnnnn

        outstr = "[MNG0000000" + str(jjul) + str(sec_in_day).zfill(5) + "GP" + str(sv).zfill(2) + "]"

        return outstr

    c = 299792458

    if interpo_30sec:
        Epoc_work = []
        Sv_work = []
        Clk_work = []

        for sv in sorted(DF["sv"].unique()):
            DFsv = DF[DF["sv"] == sv]

            Epoc_inp = np.array(conv.dt2posix(DFsv["epoch"]))
            Clk_inp = np.array(DFsv["clk"])

            Epoc_interp = np.arange(np.min(Epoc_inp), np.max(Epoc_inp), 30)
            Epoc_interp_dt = conv.posix2dt(Epoc_interp)

            I = scipy.interpolate.interp1d(Epoc_inp, Clk_inp)

            Clk_interp = I(Epoc_interp)

            Sv_work = Sv_work + [sv] * len(Epoc_interp)
            Epoc_work = Epoc_work + list(Epoc_interp_dt)
            Clk_work = Clk_work + list(Clk_interp)

    else:
        Epoc_work = DF["epoch"]
        Sv_work = DF["sv"]
        Clk_work = DF["clk"]

    DF_work = pd.DataFrame(list(zip(Epoc_work, Sv_work, Clk_work)),
                           columns=("epoch", "sv", "clk"))

    DF_work.sort_values(["epoch", "sv"], inplace=True)

    for epoc, sv, clk in zip(DF_work["epoch"], DF_work["sv"], DF_work["clk"]):
        signaletik = write_GINS_signaletic_elt_clk(epoc, sv, clk)

        str_final = " 0 0 {:}  {:+17.15e} {:+17.15e}\n".format(signaletik, clk * 10 ** -6 * c, 0)

        Fout.write(str_final)

    if not return_as_DF:
        return clk_gins_out
    else:
        return DF_work


def read_gins_multi_raw_listings(filelistin, kineorstatic='static', flh_in_rad=True):
    """
    traite une liste de listing bruts
    pour obtenir UNE SEULE timeserie
    """

    tsout = time_series.TimeSeriePoint()
    refname = 'RIEN'
    if kineorstatic == 'static':
        for filein in filelistin:
            log.info(filein)
            if not utils.check_regex(filein, 'c o n v e r g e n c e'):
                continue
            pt = read_gins(filein, kineorstatic='static', flh_in_rad=flh_in_rad)
            if refname == 'RIEN':
                refname = pt.name
            if refname != pt.name:
                log.warning("nom de stat. != reference")
            tsout.add_point(pt)
        tsout.meta_set(stat=refname)


    elif kineorstatic == 'kine':
        tsoutlis = []
        for filein in filelistin:
            log.info(filein)
            if not utils.check_regex(filein, 'c o n v e r g e n c e'):
                continue
            ts = read_gins(filein, kineorstatic='kine', flh_in_rad=flh_in_rad)
            tsoutlis.append(ts)
        tsout = time_series.merge_ts(tsoutlis)

    else:
        log.error("check kineorstatic keyword")

    tsout.sort()
    return tsout


def read_gins_multi_extracted(filelistin, flh_in_rad=True):
    """
    traite les extractions de listings
    sous la forme par ex. S<HLP>__ddhhiissxxxxxxxxxyym.HOUE
    La liste doit contenir exactement 3 fichiers (pour chacune des composantes)
    """
    tsout = time_series.TimeSeriePoint()
    if len(filelistin) != 3:
        log.error("listfilein != 3 elts")
        return None
    statnameset = list(set([f.split('.')[-1] for f in filelistin]))
    statname = statnameset[0]
    if len(statnameset) != 1:
        log.error("len(statnameset) != 1")
    fileopenedlist = [open(f, 'r+') for f in filelistin]
    coortypelist = [os.path.basename(f)[1] for f in filelistin]
    log.info("%s %s", coortypelist, filelistin)
    if 'X' in coortypelist:
        initype = 'XYZ'
        ia = coortypelist.index('X')
        ib = coortypelist.index('Y')
        ic = coortypelist.index('Z')
    elif 'p' in coortypelist:
        initype = 'FLH'
        ia = coortypelist.index('p')
        ib = coortypelist.index('L')
        ic = coortypelist.index('H')

    for lf in zip(*fileopenedlist):
        # OLD MODE : MAVAIS GESTION DE LA DATE
        #        t1 = lf[0].split()[0]
        #        t2 = lf[1].split()[0]
        #        t3 = lf[2].split()[0]
        #        if not (t1 == t2 == t3):
        #            print "WARN : read_gins_multi_extracted : not (t1 == t2 == t3)"
        #
        fields = lf[0].split()
        line = lf[0]
        blok = fields[-1]
        jour = int(blok[0:2])
        h = int(blok[2:4])
        m = int(blok[4:6])
        s = int(blok[6:8])
        yy = int(blok[-4:-2])
        # gestion des années
        if 80 < yy <= 99:
            yy = yy + 1900
        else:
            yy = yy + 2000

        # pour le mois, si > sept (9), alors lettre ...
        mm = blok[-2]

        if mm == 'O':
            mm = 10
        elif mm == 'N':
            mm = 11
        elif mm == 'D':
            mm = 12
        else:
            mm = int(mm)

        t1 = (dt.datetime(yy, mm, jour, h, m, s) + dt.timedelta(seconds=-19))

        lfiasplit = lf[ia].split()
        lfibsplit = lf[ib].split()
        lficsplit = lf[ic].split()

        A = float(lfiasplit[4])
        B = float(lfibsplit[4])
        C = float(lficsplit[4])
        sA = float(lfiasplit[5])
        sB = float(lfibsplit[5])
        sC = float(lficsplit[5])
        T = t1

        if initype == 'FLH' and flh_in_rad:
            A = np.rad2deg(A)
            B = np.rad2deg(B)
            sA = np.rad2deg(sA)
            sB = np.rad2deg(sB)

        pt = time_series.Point(A, B, C, T, initype, sA, sB, sC)

        tsout.add_point(pt)
    tsout.meta_set(stat=statname)
    tsout.sort()

    return tsout


def read_gins_double_diff(filein):
    """
    return a list of Point object for a double diff listing
    """

    if utils.grep(filein, 'c o n v e r g e n c e') == '':
        log.error('%s have no convergence, return None', filein)
        return None

    fileopened = open(filein)

    statlist = []
    rawdatalist = []
    timelist = []

    regstat = re.compile(r'[0-9]{5}[A-Z][0-9]{3}  [0-9]{7} .*$')
    regresult = re.compile(r'\[S[PLHXYZ][E ].*\]     $')

    for line in fileopened:
        # finding
        if regstat.search(line):
            stat = line.split()[3]
            statlist.append(stat)

        if regresult.search(line):
            fields = line.split()
            rawdatalist.append([float(e) for e in fields[:-2]])
            timelist.append(gins_read_time(line))

    rawdatatab = np.vstack(rawdatalist)

    Pstk = []

    for istat, stat in enumerate(statlist):
        P = time_series.Point(rawdatatab[3 * istat, 3],
                              rawdatatab[3 * istat + 1, 3],
                              rawdatatab[3 * istat + 2, 3],
                              timelist[3 * istat], 'XYZ',
                              rawdatatab[3 * istat, 4],
                              rawdatatab[3 * istat, 4],
                              rawdatatab[3 * istat, 4], name=stat)

        Pstk.append(P)

    return Pstk


def read_gins_double_diff_multi(filelistin):
    """
    return a dictionnary with station names as keys and timeseries as values
    """

    Ptsstk = []
    for filein in filelistin:
        Pstk = read_gins_double_diff(filein)
        if Pstk != None:
            Ptsstk = Ptsstk + Pstk

    statlist = set([e.name for e in Ptsstk])

    tsdico = dict()

    for stat in statlist:
        tsdico[stat] = time_series.TimeSeriePoint(stat=stat)

    for pt in Ptsstk:
        tsdico[pt.name].add_point(pt)

    for ts in tsdico.values():
        ts.sort()

    return tsdico


#   _____ ______ ______    ________ _____   ____   _____   ______ _ _
#  / ____|  ____|___  /   / /  ____|  __ \ / __ \ / ____| |  ____(_) |
# | |  __| |__     / /   / /| |__  | |__) | |  | | (___   | |__   _| | ___  ___
# | | |_ |  __|   / /   / / |  __| |  ___/| |  | |\___ \  |  __| | | |/ _ \/ __|
# | |__| | |     / /__ / /  | |____| |    | |__| |____) | | |    | | |  __/\__ \
#  \_____|_|    /_____/_/   |______|_|     \____/|_____/  |_|    |_|_|\___||___/

def read_epos_sta_kinematics(filein):
    """
    read an EPOS kinematic solutions
    """

    F = open(filein)
    Lines_4_DF_stk = []
    for l in F:
        fields = l.split()
        if l[0] != "K" and l[0] != "U" and l[0] != "X":
            continue
        if l[0] == "K" or l[0] == "U" or l[0] == "X":
            namstat = fields[2]
            numstat = int(fields[1])
            MJD_epo = float(fields[3])
            numobs = int(fields[4])

            X = float(fields[6])
            Y = float(fields[7])
            Z = float(fields[8])
            sX = float(fields[10])
            sY = float(fields[11])
            sZ = float(fields[12])

            N = float(fields[14])
            E = float(fields[15])
            U = float(fields[16])
            sN = float(fields[18])
            sE = float(fields[19])
            sU = float(fields[20])

            tup_4_df = (namstat, numstat, MJD_epo, numobs, X, Y, Z, sX, sY, sZ,
                        N, E, U, sN, sE, sU)
            Lines_4_DF_stk.append(tup_4_df)

    columns = ("site", "site_num",
               "MJD_epo", "numobs",
               "x", "y", "z", "sx", "sy", "sz",
               "N", "E", "U", "sN", "sE", "s_u")

    DFout = pd.DataFrame(Lines_4_DF_stk,
                         columns=columns)
    return DFout


def read_epos_sta_coords_mono(filein, return_df=True):
    """
    Read an GFZ EPOS's coordinate file.
    To read several files at the same time see read_epos_sta_coords_multi

    Parameters
    ----------
    filein : str
        path of the input coordinate file.
    return_df : bool, optional
        if True, returns the coordinates as a Pandas DataFrame.
        if False, return a list of GeodeZYX's point objects (advanced)
        The default is True.

    Returns
    -------
    DataFrame or List of Points
        Ouput coordinates.

    """
    # """
    # read an EPOS's YYYY_DDD_XX_sta_coordinates coordinates files
    # and return a list of Points objects

    # ... TBC ...
    # """

    from unlzw import unlzw

    if type(filein) is str:  ### case 1 : path compressed
        if filein[-2:] in (".Z"):
            with open(filein, 'rb') as fh:
                compressed_data = fh.read()
                F = unlzw(compressed_data)

        if filein[-2:] in ("gz", "GZ"):
            F = gzip.open(filein, "r+")
            F = [e.decode('utf-8') for e in F]
        else:  ### case 2 : path uncompressed
            try:
                F = open(filein, "r", encoding="ISO-8859-1")
            except:
                F = open(filein, "r")
    else:  ### case 3 : already a list of lines
        F = open(filein)

    points_list_stk = []
    lines_4_df_stk = []

    for l in F:
        fields = l.split()
        if l[0] != " ":
            continue
        if "SITE" in fields[0]:
            namestat = fields[8]
            numstat = int(fields[2])
            tecto_plate = fields[4]
            MJD_ref = int(fields[5])
            MJD_strt = int(fields[6])
            MJD_end = int(fields[7])
            MJD_mid = np.mean([MJD_strt, MJD_end])
            T = conv.numpy_dt2dt(conv.mjd2dt(MJD_mid))

        if "POS_VEL:XYZ" in fields[0]:
            X = float(fields[4])
            Y = float(fields[5])
            Z = float(fields[6])
            Vx = float(fields[7])
            Vy = float(fields[8])
            Vz = float(fields[9])

        if "SIG_PV_XYZ" in fields[0]:
            sX = float(fields[4].replace("D", "E"))
            sY = float(fields[5].replace("D", "E"))
            sZ = float(fields[6].replace("D", "E"))
            sVx = float(fields[7])
            sVy = float(fields[8])
            sVz = float(fields[9])

            #### Last useful line for the point, store it
            if not return_df:
                point = time_series.Point(X, Y, Z, T, "XYZ", sX, sY, sZ, name=namestat)
                point.anex["Vx"] = sVx
                point.anex["Vy"] = sVy
                point.anex["Vz"] = sVz
                points_list_stk.append(point)

            #### And store for the DataFrame
            else:
                tup_4_DF = (namestat, numstat, tecto_plate,
                            conv.mjd2dt(MJD_strt),
                            MJD_ref, MJD_strt, MJD_end,
                            X, Y, Z, sX, sY, sZ,
                            Vx, Vy, Vz, sVx, sVy, sVz)

                lines_4_df_stk.append(tup_4_DF)

    if return_df:
        columns = ("site", "site_num", "tecto_plate", "epoch",
                   "MJD_ref", "MJD_start", "MJD_end",
                   "x", "y", "z", "sx", "sy", "sz",
                   "Vx", "Vy", "Vz", "sVx", "sVy", "sVz")

        DFout = pd.DataFrame(lines_4_df_stk,
                             columns=columns)

        return DFout
    else:
        return points_list_stk


def read_epos_sta_coords_multi(filein_list, output_type="DataFrame"):
    """
    Read several GFZ EPOS's coordinate files.

    Parameters
    ----------
    filein_list : list
        list of input coordinate files inputs.
    output_type : str, optional
        "DataFrame": returns a Pandas DataFrame containing the coordinates
        "TSobjects": returns a dictionary of GeodeZYX's TimeSeries objects (advanced)
        The default is "DataFrame".

    Returns
    -------
    OUT : DataFrame or dict
        See "output_type" input parameter.

    """

    if output_type == "TSobjects":
        OUT = read_epos_sta_coords_multi_legacy(filein_list, return_dict=True)
    elif output_type == "DataFrame":
        DFfil_stk = []
        for fil in filein_list:
            DFfil = read_epos_sta_coords_mono(fil, return_df=True)
            DFfil_stk.append(DFfil)
        DFall = pd.concat(DFfil_stk)
        DFall.reset_index(inplace=True, drop=True)
        OUT = DFall

    return OUT


def read_epos_sta_coords_multi_legacy(filein_list, return_dict=True):
    """
    Read several GFZ EPOS's coordinate files.
    Legacy version
    
    Parameters
    ----------
    filein_list : list
        list of input coordinate files inputs.
    return_dict : bool, optional
        True: returns a dictionary of GeodeZYX's TimeSeries objects 
        "TSobjects": returns a list of GeodeZYX's TimeSeries objects 
        The default is True.

    Returns
    -------
    OUT : dict or list
        See "return_dict" input parameter.
    """

    filein_list = sorted(filein_list)
    Points_list = []
    statname_stk = []

    for fil in filein_list:
        Points_daily_list = read_epos_sta_coords_mono(fil, return_df=False)
        Points_list = Points_list + Points_daily_list
        statname_stk = statname_stk + [e.name for e in Points_daily_list]

    statname_uniq = sorted(list(set(statname_stk)))

    ts_dict = dict()

    for point in Points_list:
        if not point.name in ts_dict.keys():
            ts_dict[point.name] = time_series.TimeSeriePoint(stat=point.name)
        ts_dict[point.name].add_point(point)

    if return_dict:
        return ts_dict
    else:
        ts_list = []
        for k, val in ts_dict.items():
            ts_list.append(val)
        return ts_list


def read_epos_slv_times(p, convert_to_time=False):
    """
    convert_to_time : divide by the speed of light to get time-homogene values.
    Values in meter instead
    If convert_to_time : time in sec
    """

    L = utils.extract_text_between_elements_2(p, r"\+sum_times/estimates",
                                              r"\-sum_times/estimates")

    Lgood_stat = []
    Lgood_sat = []

    for l in L[1:-1]:
        if "EPOCHE" in l:
            cur_epoc_line = l
            cur_epoc_f = cur_epoc_line.split()
            cur_epoc = conv.mjd2dt(int(cur_epoc_f[1])) + dt.timedelta(seconds=int(86400 * float(cur_epoc_f[2])))

        if re.match("^   [0-9]{4}.*", l):
            Lgood_stat.append([cur_epoc] + [float(e) for e in l.split()])

        if re.match("^ [A-Z][0-9]{2}.*", l):
            e = l.split()
            Lgood_sat.append([cur_epoc] + [e[0], float(e[1]), float(e[2])])

    ### stations
    DF_stat = pd.DataFrame(Lgood_stat, columns=["epoch", "stat", "offset", "offset_sig"])
    DF_stat["stat"] = DF_stat["stat"].astype('int')
    if convert_to_time:
        DF_stat[["offset", "offset_sig"]] = DF_stat[["offset", "offset_sig"]] / 299792458.

    ### satellites
    DF_sat = pd.DataFrame(Lgood_sat, columns=["epoch", "sat", "offset", "offset_sig"])
    if convert_to_time:
        DF_sat[["offset", "offset_sig"]] = DF_sat[["offset", "offset_sig"]] / 299792458.

    return DF_stat, DF_sat


def read_epos_tim(tim_file_in, convert_to_sec=False):
    """
    results in microsec
    """
    F = open(tim_file_in)

    head_stop = False

    if convert_to_sec:
        koef = 10 ** -6
    else:
        koef = 1.

    Val_stk = []
    for l in F:
        if re.match(r'^\*  [0-9]{4} *([0-9]{1,2} *){4}', l):
            head_stop = True
            epoc = conv.datetime_improved(*l[3:30].split())
        if head_stop and re.match('[A-Z][0-9]{2}.* [0-9]*', l):
            val = l.split()
            val[1] = float(val[1]) * koef
            val.insert(0, epoc)

            Val_stk.append(val)

    DF = pd.DataFrame(Val_stk, columns=["epoch", "sat", "offset"])

    return DF


def read_nevada(filein, input_coords="enu"):
    """
    input_coords="enu" or "xyz"
    """

    tsout = time_series.TimeSeriePoint()

    envfile = open(filein)

    if input_coords == "enu":
        for l in envfile:
            f = l.split()

            if "site YYMMMDD" in l:
                continue
            if len(l) == 0:
                continue

            stat = f[0]

            T = conv.year_decimal2dt(float(f[2]))

            N = float(f[10])
            E = float(f[8])
            U = float(f[12])

            sN = float(f[15])
            sE = float(f[14])
            sU = float(f[16])

            point = time_series.Point(E, N, U, T, 'ENU', sE, sN, sU)

            # tsout.refENU = time_series.Point()

            tsout.boolENU = True
            tsout.add_point(point)

    if input_coords == "xyz":
        for l in envfile:
            f = l.split()

            if "site YYMMMDD" in l:
                continue
            if len(l) == 0:
                continue

            stat = f[0]

            T = conv.year_decimal2dt(float(f[2]))

            X = float(f[3])
            Y = float(f[4])
            Z = float(f[5])

            sX = float(f[6])
            sY = float(f[7])
            sZ = float(f[8])

            point = time_series.Point(X, Y, Z, T, 'XYZ', sX, sY, sZ)

            point.anex['Rxy'] = float(f[9])
            point.anex['Rxz'] = float(f[10])
            point.anex['Ryz'] = float(f[11])

            tsout.add_point(point)

    tsout.stat = stat

    return tsout


def read_IGS_coords(filein, initype='auto'):
    Tstk, Astk, Bstk, Cstk = [], [], [], []
    tsout = time_series.TimeSeriePoint()
    for l in open(filein):
        f = l.split()

        if initype == 'auto':
            if "plh" in os.path.basename(filein):
                initype = 'FLH'
            elif "xyz" in os.path.basename(filein):
                initype = 'XYZ'

        T = conv.mjd2dt(float(f[3]))
        A = float(f[6])
        B = float(f[7])
        C = float(f[8])
        sA = float(f[9])
        sB = float(f[10])
        sC = float(f[11])

        initype = "FLH"

        pt = time_series.Point(A, B, C, T, initype, sA, sB, sC, f[0])
        tsout.add_point(pt)

    tsout.meta_set(filein, f[0])
    return tsout


def sorting_a_calais_file(openedfile):
    openedfile.seek(0)
    T, A, sA, STAT = [], [], [], []
    for l in openedfile:
        f = l.split()
        T.append(float(f[0]))
        A.append(float(f[1]))
        sA.append(float(f[2]))
        STAT.append(str(f[3]))

    DATA = [T, A, sA, STAT]
    DATA2 = utils.sort_table(DATA, 0)

    return DATA2


def read_calais(filelist):
    """ filelistin est une liste de 3 fichier E N & U """

    if filelist == []:
        log.warning('files list empty , exiting ...')
        return None
    filelist.sort()

    fileopenedlist = [open(f) for f in filelist]

    sorted_data_lis = [sorting_a_calais_file(fil) for fil in fileopenedlist]

    statnameset = list(set([f.split('.')[0] for f in filelist]))
    statname = os.path.basename(statnameset[0])

    # LOADING ALL DATA IN A BIG MATRIX
    bigT = np.array(sorted(set(np.hstack([np.array(sd[0]) for sd in sorted_data_lis]))))
    bigDATA = np.empty((len(bigT), 3))
    bigDATA.fill(np.nan)
    bigDATAsigma = np.empty((len(bigT), 3))
    bigDATAsigma.fill(np.nan)

    for i, bt in enumerate(bigT):
        for j, data in enumerate(sorted_data_lis):
            for t, d, sd, stat in zip(*data):
                if t == bt:
                    bigDATA[i, j] = d
                    bigDATAsigma[i, j] = sd

    DATA = np.hstack((bigDATA / 1000., bigDATAsigma / 100.))

    ptslist = []

    # MAKING POINTS
    for i in range(DATA.shape[0]):
        pt = time_series.Point(conv.year_decimal2dt(bigT[i])
                               , 'ENU', DATA[i, 3], DATA[i, 4], DATA[i, 5])
        ptslist.append(pt)

    tsout = time_series.TimeSeriePoint()

    for pt in ptslist:
        tsout.add_point(pt)

    # FINDING DISCONT
    # finding discont directly in files

    # Finding the composant with max of data
    lendata = [len(sd[0]) for sd in sorted_data_lis]
    ii = lendata.index(max(lendata))

    discont = []
    T = sorted_data_lis[ii][0]
    STAT = sorted_data_lis[ii][-1]
    for i in range(len(STAT) - 1):
        if STAT[i + 1] != STAT[i]:
            discont.append(T[i + 1])

    tsout.set_discont(discont)
    tsout.meta_set(stat=statname)
    tsout.boolENU = True
    tsout.sort()

    return tsout


def read_renag_synthetic(filein, discont_file_in=None):
    tsout = time_series.TimeSeriePoint()

    fil = open(filein)

    for l in fil:

        if l[0] == "#":
            continue

        f = l.split()

        T = conv.year_decimal2dt(float(f[0]))

        N = float(f[1])
        E = float(f[2])
        U = float(f[3])

        sN = float(f[4])
        sE = float(f[5])
        sU = float(f[6])

        point = time_series.Point(E, N, U, T, 'ENU', sE, sN, sU)

        # tsout.refENU = time_series.Point()

        tsout.boolENU = True
        tsout.add_point(point)

    if discont_file_in:
        DiscontInp = open(discont_file_in)

        Discont = []
        for l in DiscontInp:
            if l[0] == "#":
                continue
            f = l.split()
            try:
                Discont.append(conv.doy2dt(int(f[0]), int(f[1])))
            except:
                log.warning("something went wrong during discont. file reading")
                pass

        Discont = sorted(Discont)
        tsout.set_discont(Discont)

    stat_name = os.path.basename(filein).split(".")[0].split(".")[0]
    tsout.meta_set(path=filein, stat=stat_name, name=stat_name)

    return tsout


def read_jump_file(filein, returned_events=('S', 'E', 'D')):
    """
    From a "Jump" File (p. Sakic internal file)
    Return a dictionnairy with events

    Parameters
    ----------
    filein : str
        path of the Jump File

    returned_events : tuple or list
        contains the inital letter of the event type which will be stored in the dico
        (See below)

    Returns
    -------
    jump_dico : dict of dict of datetime
        outputed events, in the form jump_dico["STAT"]["L"] = datetime
        where L is the inital letter of the event type

    Note
    ----
    A jump file contains infos like this :

    #>>> STAT S 2000 001
    #>>> STAT E 2001 001
    #>>> STAT D 2000 06 01

    it can manage YEAR DOY or YEAR MM DD or DECIMAL YEAR

    A non-blank 1st column is a commented line

    After a #, it is a commentary

    event type letters :
        S : Start

        E : End

        D : Discontinuity

    """
    import pytz

    F = open(filein)
    jump_dico = dict()
    for l in F:
        l = l.split("#")[0]
        f = l.split()

        ## Skip comment lines
        if (not f) or (l[0] != " ") or ("#" in l):
            continue
        else:
            stat = f[0]
            event = f[1]
            ## Create the key for the station
            if not f[0] in jump_dico.keys():
                jump_dico[stat] = dict()
                for rtn_evt in returned_events:
                    jump_dico[stat][rtn_evt] = []

            ## Fill the dico
            if event in returned_events:
                if len(f) == 4:  ### DOY
                    date = conv.doy2dt(int(f[2]), int(f[3]))
                elif len(f) == 5:  ### YYYY MM DD
                    date = dt.datetime(*[int(e) for e in f[2:]])
                elif len(f) == 3:  ### DECIMAL YEAR
                    date = conv.year_decimal2dt(float(f[2]))

                date_tz = date.replace(tzinfo=pytz.UTC)
                jump_dico[stat][event].append(date_tz)
    return jump_dico


def read_nav_step1_geodesea(filein):
    M = np.loadtxt(filein)
    tsout = time_series.TimeSeriePoint()

    for m in M:
        pt = time_series.Point(np.rad2deg(m[1]), np.rad2deg(m[2]), m[3], m[0], initype='FLH')
        tsout.add_point(pt)

    return tsout


#  _   _ _____   _____          _   _   ______ _ _
# | \ | |  __ \ / ____|   /\   | \ | | |  ____(_) |
# |  \| | |__) | |       /  \  |  \| | | |__   _| | ___  ___
# | . ` |  _  /| |      / /\ \ | . ` | |  __| | | |/ _ \/ __|
# | |\  | | \ \| |____ / ____ \| |\  | | |    | | |  __/\__ \
# |_| \_|_|  \_\\_____/_/    \_\_| \_| |_|    |_|_|\___||___/

def read_nrcan_csv(filein, associated_ps_file='', statname=''):
    """
    associated_ps_file is highly recommanded
    because of the time managing

    WARN : Must be avoided b/c of the weak decimal precision of the angles !!!
    """

    if statname == '':
        statname = os.path.basename(filein)[0:4]

    pdcsv = pd.read_csv(filein)

    F = np.array(pdcsv['latitude_degre_decimal'])
    L = np.array(pdcsv['longitude_degre_decimal'])
    H = np.array(pdcsv['hauteur_ellipsoidale_m'])
    heure = np.array(pdcsv['heure_decimal'])
    doy = np.array(pdcsv['jour_de_l_annee'])
    year = np.array(pdcsv['annee'])

    T = conv.doy2dt(year, doy, heure)

    if associated_ps_file != '':
        T = []
        for l in open(associated_ps_file):
            if "BWD" in l:
                f = l.split()
                t = conv.date_string_2_dt(f[4] + ' ' + f[5])
                T.append(t)
        if statname == '':
            statname = f[2]

    tsout = time_series.TimeSeriePoint()
    for f, l, h, t in zip(F, L, H, T):
        point = time_series.Point(f, l, h, t, 'FLH', name=statname)
        tsout.add_point(point)

    tsout.meta_set(filein, statname)

    return tsout


def read_nrcan_pos(filein):
    """
    .pos file are more precise than .csv, should be used !
    """
    tsout = time_series.TimeSeriePoint()
    start_read = False

    for l in open(filein):
        if l[0:3] == 'DIR':
            start_read = True
            lhead = l.split()
            i_lat_d = lhead.index('LATDD')
            i_lat_m = lhead.index('LATMN')
            i_lat_s = lhead.index('LATSS')

            i_lon_d = lhead.index('LONDD')
            i_lon_m = lhead.index('LONMN')
            i_lon_s = lhead.index('LONSS')

            i_h = lhead.index('HGT(m)')

            i_slat = lhead.index('SDLAT(95%)')
            i_slon = lhead.index('SDLON(95%)')
            i_sh = lhead.index('SDHGT(95%)')

            continue
        elif not start_read:
            continue
        else:
            f = l.split()
            lat = (np.abs(float(f[i_lat_d])) + 1 / 60. * float(f[i_lat_m]) + 1 / 3600. * float(f[i_lat_s])) * np.sign(
                float(f[i_lat_d]))
            lon = (np.abs(float(f[i_lon_d])) + 1 / 60. * float(f[i_lon_m]) + 1 / 3600. * float(f[i_lon_s])) * np.sign(
                float(f[i_lon_d]))
            h = float(f[i_h])

            ### old and useless conversion (2021-01)
            # sE = float(f[15])
            # sN = float(f[16])
            # sU = float(f[17])
            # slat , slon , sh = conv.sigma_enu2geo(lat,lon,h,sE,sN,sU)

            slat, slon, sh = float(f[i_slat]), float(f[i_slon]), float(f[i_sh])

            t = conv.date_string_2_dt(f[4] + ' ' + f[5])

            pt = time_series.Point(lat, lon, h, t, 'FLH', slat, slon, sh, name=f[2])
            tsout.add_point(pt)

    tsout.meta_set(filein, f[2])
    return tsout


def read_qinsy(filein, yy, mm, dd):
    reader = pd.read_csv(open(filein))
    T = [dateutil.parser.parse(e).replace(year=yy, month=mm, day=dd) + dt.timedelta(seconds=dUTCGPS) for e in
         list(reader.icol(0))]
    (X, Y, Z) = np.array(conv.geo2xyz(reader.icol(12), reader.icol(13), reader.icol(14)))
    # sX,sY,sZ = [] , [] , []
    initype = 'XYZ'
    tsout = time_series.TimeSeriePoint()
    for i in range(len(T)):
        point = time_series.Point(X[i], Y[i], Z[i], T[i], initype, sA=np.nan, sB=np.nan, sC=np.nan)
        tsout.add_point(point)
    tsout.meta_set(filein)
    return tsout


def read_sonardyne_posi(filein, dUTCGPS):
    reader = pd.read_csv(open(filein), skip_footer=1)
    T = [dateutil.parser.parse(e) + dt.timedelta(seconds=dUTCGPS) for e in list(reader['UTCTime'])]
    (L, F, H) = (reader['Longitude'], reader['Latitude'], reader['Altitude'])
    sX, sY, sZ = [], [], []
    initype = 'FLH'
    tsout = time_series.TimeSeriePoint()
    for i in range(len(T)):
        point = time_series.Point(F[i], L[i], H[i], T[i], initype, sA=np.nan, sB=np.nan, sC=np.nan)
        tsout.add_point(point)

    tsout.meta_set(filein)
    return tsout


def read_pbo_pos(filein):
    filobj = open(filein, 'r')
    tsout = time_series.TimeSeriePoint()
    header = True
    for line in filobj:
        if not header:
            f = line.split()
            f2 = [float(e) for e in f[:-1]]
            t = dt.datetime(int(f[0][0:4]), int(f[0][4:6]), int(f[0][6:]), int(f[1][0:2]), int(f[1][2:4]),
                            int(f[1][4:]))
            pt = time_series.Point(f2[3], f2[4], f2[5], t, 'XYZ', f2[6], f2[7], f2[8])
            pt.FLHset(f2[12], f2[13], f2[14])
            pt.ENUset(f2[16], f2[15], f2[17], f2[19], f2[18], f2[20])
            pt.anex['Rxy'] = f2[9]
            pt.anex['Rxz'] = f2[10]
            pt.anex['Ryz'] = f2[11]
            pt.anex['Rne'] = f2[-4]
            pt.anex['Rnu'] = f2[-3]
            pt.anex['Reu'] = f2[-2]
            tsout.add_point(pt)
        if line[0] == '*':
            header = False
    tsout.boolENU = True
    tsout.meta_set(filein)
    tsout.stat = os.path.basename(filein)[0:4]
    return tsout


def read_sonardyne_attitude(filein):
    reader = pd.read_csv(open(filein), skip_footer=1)

    T = [dateutil.parser.parse(e) + dt.timedelta(seconds=dUTCGPS) for e in list(reader['PTPTime'])]

    pitch = np.array(reader['Pitch'])
    roll = np.array(reader['Roll'])
    head = np.array(reader['Heading'])

    initype = 'FLH'

    AHRS = list(reader['AHRS Id'])
    nbunit = reffram.guess_seq_len(AHRS)
    log.info("nbre de devices ID : %s", nbunit)
    AHRS = reader['AHRS Id']

    tsout_list = []

    for n in range(nbunit):
        tsout = TimeSerieObs()
        tsout.typeobs = 'RPY'
        tsout_list.append(tsout)

    i = 0
    for p, r, h, t, a in zip(roll, pitch, head, T, AHRS):
        att = Attitude(p, r, h, t, devID=a)
        n = np.mod(i, nbunit)
        tsout_list[n].add_obs(att)
        i = i + 1

    for ts in tsout_list:
        ts.meta_set(filein, devID=list(set([e.devID for e in ts.obs])))
        ts.interp_set()

    return tsout_list


def interp_sndy_SYS_UTC(time_conv_file_in):
    timeConv = utils.read_mat_file(time_conv_file_in)
    timeSYS = timeConv[0, :]
    timeUTC = timeConv[1, :]
    IntSYSUTC = scipy.interpolate.interp1d(timeSYS, timeUTC, kind='slinear',
                                           bounds_error=0)
    TSout = time_series.TimeSerieObs(time_conv_file_in)

    return IntSYSUTC


def read_sndy_mat_att(filein, IntSYSUTCin=None):
    if IntSYSUTCin == None:
        log.warning("No Interpolator")
    attmat = utils.read_mat_file(filein)
    Tsys_att = attmat[0, :]
    Tposix_att = IntSYSUTCin(Tsys_att)
    #    Tdt_att = np.array(conv.posix2dt(Tposix_att))
    roll = attmat[1, :]
    pitch = attmat[2, :]
    head = attmat[3, :]

    TSout = time_series.TimeSerieObs('RPY', filein)

    for r, p, h, t in zip(roll, pitch, head, Tposix_att):
        att = time_series.Attitude(r, p, h, t)
        TSout.add_obs(att)
    return TSout


def read_sndy_mat_nav(filein, IntSYSUTCin=None):
    if IntSYSUTCin == None:
        log.warning("No Interpolator")
    navmat = utils.read_mat_file(filein)
    Tsys_nav = navmat[0, :]
    Tposix_nav = IntSYSUTCin(Tsys_nav)
    #    Tdt_att = np.array(conv.posix2dt(Tposix_att))
    lat = navmat[2, :]
    lon = navmat[3, :]
    h = navmat[4, :]

    TSout = time_series.TimeSeriePoint()

    for f, l, h, t in zip(lat, lon, h, Tposix_nav):
        pt = time_series.Point(f, l, h, t, initype='FLH')
        TSout.add_point(pt)
    return TSout


def read_hector_neu(filein):
    log.warning("XYZ/FLH conversion not implemented")
    M = np.loadtxt(filein)
    stat = utils.grep(filein, 'Site :', only_first_occur=True).split()[3]
    tsout = time_series.ts_from_list(M[:, 2], M[:, 1], M[:, 3],
                                     conv.year_decimal2dt(M[:, 0]),
                                     'ENU',
                                     M[:, 4], M[:, 5], M[:, 6],
                                     stat=stat, name=stat)

    return tsout


#  _______ _    _  _____     _______ _____   ____   ____  _____   _____   ______ _ _
# |__   __| |  | |/ ____|   / / ____|  __ \ / __ \ / __ \|  __ \ / ____| |  ____(_) |
#    | |  | |  | | |  __   / / |  __| |__) | |  | | |  | | |__) | (___   | |__   _| | ___  ___
#    | |  | |  | | | |_ | / /| | |_ |  _  /| |  | | |  | |  ___/ \___ \  |  __| | | |/ _ \/ __|
#    | |  | |__| | |__| |/ / | |__| | | \ \| |__| | |__| | |     ____) | | |    | | |  __/\__ \
#    |_|   \____/ \_____/_/   \_____|_|  \_\\____/ \____/|_|    |_____/  |_|    |_|_|\___||___/


def read_groops_position(Filesin):
    if not utils.is_iterable(Filesin):
        Filesin = [Filesin]

    Statnames = list(set([os.path.basename(f)[-8:-4] for f in Filesin]))

    if len(Statnames) > 1:
        log.warn("several stations for the same TimeSerie!:" + str(Statnames))

    statname = Statnames[0]

    tsout = time_series.TimeSeriePoint()

    for filein in Filesin:
        DF = pd.read_csv(filein, skiprows=6, header=None, sep=r'\s+')
        T = conv.dt2posix(conv.mjd2dt(DF[0].values))
        X, Y, Z = DF[1], DF[2], DF[3]

        for t, x, y, z in zip(T, X, Y, Z):
            point = time_series.Point(x, y, z, t, 'XYZ', name=statname)
            tsout.add_point(point)

    tsout.meta_set(stat=statname)
    tsout.sort()

    return tsout


def _pride_pppar_end_header(filein):
    colheader = 0
    stat = "XXXX"

    with open(filein) as F:
        L = F.readlines()

    for i, l in enumerate(L):
        if "STATION" in l:
            stat = l.split()[0]

        if "END OF HEADER" in l:
            colheader = i + 1
            break
    return colheader, stat


def read_pride_pppar_pos_mono(filein):
    colheader, stat_header = _pride_pppar_end_header(filein)

    df = pd.read_csv(filein, skiprows=colheader + 1,
                     # delim_whitespace=True,
                     sep=r'\s?\*?\s+',
                     engine='python',
                     header=None)

    df.columns = ['stat', 'Mjd', 'X', 'Y', 'Z', 'Sx', 'Sy', 'Sz',
                  'Rxy', 'Rxz', 'Ryz', 'Sig0', 'Nobs']

    df = df.squeeze()

    T = conv.dt2posix(conv.mjd2dt(df['Mjd']))

    fuv = df['Sig0'] ** 2  # variance of unit weight

    anex = dict()
    anex['sdXY'] = df['Rxy'] * fuv
    anex['sdXZ'] = df['Rxz'] * fuv
    anex['sdYZ'] = df['Ryz'] * fuv

    # not sure for the sigma computation
    pt = time_series.Point(df['X'], df['Y'], df['Z'], T, 'XYZ',
                           np.sqrt(df['Sx']) * fuv,
                           np.sqrt(df['Sy']) * fuv,
                           np.sqrt(df['Sz']) * fuv,
                           name=df['stat'], anex=anex)

    return pt


def read_pride_pppar_pos(files_list_in):
    tsout = time_series.TimeSeriePoint()

    for file in files_list_in:
        try:
            pt = read_pride_pppar_pos_mono(file)
        except pd.errors.EmptyDataError as e:
            log.error("%s, %s skipped", e, file)
            continue

        tsout.add_point(pt)
    tsout.meta_set(stat=pt.name)
    tsout.sort()

    return tsout


def read_pride_pppar_kin(filein):
    stat = "XXXX"

    colheader, stat = _pride_pppar_end_header(filein)

    df = pd.read_csv(filein, skiprows=colheader + 1,
                     # delim_whitespace=True,
                     sep=r'\s?\*?\s+',
                     engine='python',
                     header=None)

    t_arr = conv.mjd2dt(df[0]) + df[1].apply(lambda x: dt.timedelta(seconds=x))

    tsout = time_series.TimeSeriePoint()

    tsout = time_series.ts_from_list(df[2].values,
                                     df[3].values,
                                     df[4].values,
                                     t_arr, 'XYZ',
                                     stat=stat,
                                     name=stat)

    return tsout


def read_webobs(filein, typein="txt",
                coordtreat=False,
                dropna=False):
    if coordtreat:
        lbda_colname = lambda c: c + "_treat"
    else:
        lbda_colname = lambda c: c + "ern" if c != "Up" else "Up"

    if typein == "txt":
        header = utils.grep(filein, "#")[-1][1:].strip().split()
        DF = pd.read_csv(filein, sep=" ", comment='#', names=header,
                         on_bad_lines="warn")
        unit_suffix = "(m)"
        ### Time  conversion
        DFtime = DF[["yyyy", "mm", "dd", "HH", "MM", "SS"]].copy()
        DFtime.columns = ['year', 'month', 'day', 'h', 'm', 's']
        DF["T"] = pd.to_datetime(DFtime)

    elif typein == "csv":
        DF = pd.read_csv(filein, sep=";",
                         on_bad_lines="warn")
        unit_suffix = ""
        ### Time  conversion
        DFtimedelta = pd.to_timedelta(DF.HH * 3600 + DF.MM * 60 + DF.SS,
                                      unit="S")
        DF["T"] = pd.to_datetime(DF["yyyy-mm-dd"]) + DFtimedelta

    if dropna:
        DF = DF.dropna()

    tsout = time_series.TimeSeriePoint()

    T = conv.dt2posix(DF['T'].values)
    A = DF[lbda_colname("East") + unit_suffix].values
    B = DF[lbda_colname("North") + unit_suffix].values
    C = DF[lbda_colname("Up") + unit_suffix].values

    tsout.from_list(T, A, B, C, coortype='UTM')

    return tsout


def read_spotgins_masterfile(master_inp):
    return pd.read_fwf(master_inp, infer_nrows=11, skiprows=5, comment="#")

#  ______                _   _                _____                                         _
# |  ____|              | | (_)              / ____|                                       | |
# | |__ _   _ _ __   ___| |_ _  ___  _ __   | |  __ _ __ __ ___   _____ _   _  __ _ _ __ __| |
# |  __| | | | '_ \ / __| __| |/ _ \| '_ \  | | |_ | '__/ _` \ \ / / _ \ | | |/ _` | '__/ _` |
# | |  | |_| | | | | (__| |_| | (_) | | | | | |__| | | | (_| |\ v /  __/ |_| | (_| | | | (_| |
# |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|  \_____|_|  \__,_| \_/ \___|\__, |\__,_|_|  \__,_|
#                                                                        __/ |
#                                                                       |___/


# def read_gins_kinematic(filein):
#
#    ''' retourne une TSPoint '''
#
#    tsout = time_series.TimeSeriePoint()
#
#    X,Y,Z = 0,0,0
#    Tx , Ty , Tz, T = 111,222,333,0
#    sX,sY,sZ = 0,0,0
#
#    for line in open(filein):
#
#        if re.compile('\[S[XYZ] .*\]$').search(line):
#            fields = line.split()
#
#            if (float(fields[2]) == 0):
#                continue
#
#            if (fields[0] == 'stations'):
#                continue
#            # securité pour les lignes du type
#            #  ------------------------------------------------------------------------------------------------------
#            # stations      175  -0.574353783560234E+07  +/-   0.000000000000000E+00   1  [SX  1212001892701M005071]
#
#            jour = int(line[108:110])
#            h = int(line[110:112])
#            m = int(line[112:114])
#            s = int(line[114:116])
#            yy = int(line[125:127]) + 2000
#            mm = int(line[127])
#
#            if line[105] == 'X':
#                Tx = (dt.datetime(yy,mm,jour,h,m,s) + dt.timedelta(seconds=-19))
#                X = (float(fields[3]))
#                sX = (float(fields[4]))
#
#            if  line[105] == 'Y':
#                Ty = (dt.datetime(yy,mm,jour,h,m,s) + dt.timedelta(seconds=-19))
#                Y = (float(fields[3]))
#                sY = (float(fields[4]))
#
#            if  line[105] == 'Z':
#                Tz = (dt.datetime(yy,mm,jour,h,m,s) + dt.timedelta(seconds=-19))
#                Z = (float(fields[3]))
#                sZ = (float(fields[4]))
#
#
#            if  Tx == Ty == Tz :
#                T = Tx
#                point = time_series.Point(X,Y,Z,T,'XYZ',sX,sY,sZ)
#                tsout.add_point(point)
#                Tx = 111
#                Ty = 222
#                Tz = 333
#
#    tsout.meta_set(filein)
#
#    return tsout
#
#
# def read_gins_static_solo(filein):
#
#    X,Y,Z = 0,0,0
#    Tx , Ty , Tz, T = 111,222,333,0
#    sX,sY,sZ = 0,0,0
#    namestat='NULL'
#
#    fileopened = open(filein)
#
#    for line in fileopened:
#
#        if re.compile('__Nom__').search(line):
#            namestat = next(fileopened).split()[3]
#
#        # Pour le static, il y a des blancs en fin de ligne ...
#        if re.compile('\[S[XYZ] .*\]     $').search(line):
#
#            fields = line.split()
#
#            if (float(fields[2]) == 0):
#                continue
#
#            if (fields[0] == 'stations'):
#                continue
#            # securité pour les lignes du type
#            #  ------------------------------------------------------------------------------------------------------
#            # stations      175  -0.574353783560234E+07  +/-   0.000000000000000E+00   1  [SX  1212001892701M005071]
#
#            jour = int(line[108:110])
#            h = int(line[110:112])
#            m = int(line[112:114])
#            s = int(line[114:116])
#            yy = int(line[125:127])
#
#            # gestion des annÃ©es
#            if 80 < yy <= 99:
#                yy = yy + 1900
#            else:
#                yy = yy + 2000
#
#            # pour le mois, si > sept (9), alors lettre ...
#            mm = line[127]
#
#            if mm == 'O':
#                mm = 10
#            elif mm == 'N':
#                mm = 11
#            elif mm == 'D':
#                mm = 12
#            else:
#                mm = int(mm)
#
#            if line[105] == 'X':
#                Tx = (dt.datetime(yy,mm,jour,h,m,s) + dt.timedelta(seconds=-19))
#                X = (float(fields[3]))
#                sX = (float(fields[4]))
#
#            if  line[105] == 'Y':
#                Ty = (dt.datetime(yy,mm,jour,h,m,s) + dt.timedelta(seconds=-19))
#                Y = (float(fields[3]))
#                sY = (float(fields[4]))
#
#            if  line[105] == 'Z':
#                Tz = (dt.datetime(yy,mm,jour,h,m,s) + dt.timedelta(seconds=-19))
#                Z = (float(fields[3]))
#                sZ = (float(fields[4]))
#
#
#            if  Tx == Ty == Tz :
#                T = Tx
#                point = time_series.Point(X,Y,Z,T,'XYZ',sX,sY,sZ,name=namestat)
#
#                Tx = 111
#                Ty = 222
#                Tz = 333
#
#                return point

# read_gins_static_solo('/media/DDannex/GINS_AKRIM/listing/listing_NRMD/DIR_sortie_2007_20830_nrmd.140513_060821.gins')
